# -----------------------------------------------------------------------------
# Copyright (C) 2024-2025 L. Meunier

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
from scipy import stats
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_score
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import StandardScaler
from statsmodels.formula.api import ols

from ..utils import round_scientific_notation, truncate_name


class CohortMetric:
    """
    A class for performing quality control (QC) on cohort datasets, including PCA analysis,
    batch effect correction assessment.

    Attributes
    ----------
    clinical_df : pd.DataFrame
        Clinical data frame containing patient/sample metadata.
    data_expression_df : pd.DataFrame
        Gene expression data frame after batch effect correction.
    data_expression_df_before : pd.DataFrame
        Gene expression data frame before batch effect correction.
    covariates : list
        List of covariate names used for analysis.
    mixed_datasets : list
        List of datasets identified as mixed datasets based on covariate combinations.
    n_components : int
        Number of principal components to retain.
    n_neighbors : int
        Number of nearest neighbors to consider in the entropy calculation.

    Methods
    -------
    identify_mixed_datasets()
        Identifies datasets that contain several combinations of covariates.
    pca_analysis()
        Performs PCA and computes associations between PCs and clinical annotations.
    cohort_summary()
        Generates a summary of the cohort.
    quantify_correction_effect()
        Computes a metric quantifying the effect of batch correction.
    compare_sample_distribution_by_covariates(combination)
        Compares sample distribution by covariates.
    compare_sample_distribution_total()
        Compares overall sample distribution before and after correction.
    silhouette_score()
        Computes the Silhouette Score for batch mixing before and after correction.
    compute_entropy(data_expression_df)
        Computes entropy of batch mixing.
    entropy_batch_mixing()
        Computes entropy of batch mixing before and after correction.
    summarize_and_compare_mixed_datasets()
        Summarizes and compares mixed datasets.
    process()
        Runs all QC checks.
    """

    def __init__(
        self,
        clinical_df: pd.DataFrame,
        batch_column: str,
        data_expression_df: pd.DataFrame,
        data_expression_df_before: pd.DataFrame = None,
        covariates: list = None,
        clinical_columns_of_interest: list = None,
        n_components: int = 10,
        n_neighbors: int = 5,
    ) -> None:
        """
        Initialize the CohortMetric object with clinical and expression data.

        Parameters
        ----------
        clinical_df : pandas.DataFrame
            Clinical data containing patient/sample metadata.
        data_expression_df : pandas.DataFrame
            Gene expression data after batch effect correction.
        data_expression_df_before : pandas.DataFrame
            Gene expression data before batch effect correction.
        covariates : list, optional
            List of covariate names used for analysis, by default None.
        clinical_columns_of_interest : list, optional
            List of clinical columns of interest to include in the analysis, by default None.
        n_components : int, optional
            Number of principal components to retain, by default 10.
        n_neighbors : int, optional
            Number of nearest neighbors to consider in the entropy calculation, by default 5.

        Raises
        ------
        ValueError
            If any specified covariates are not present in the clinical data frame.
        """
        common_samples = np.intersect1d(data_expression_df.columns, clinical_df.index)
        if clinical_columns_of_interest is None:
            clinical_columns_of_interest = clinical_df.columns.tolist()
        self.clinical_df = clinical_df.loc[common_samples, clinical_columns_of_interest]

        if covariates is None:
            covariates = []
        # Check if covariates are in clinical_df
        missing_covariates = [
            covariate
            for covariate in covariates
            if covariate not in clinical_df.columns
        ]
        if len(missing_covariates) > 0:
            raise ValueError(
                f"Covariates {', '.join(missing_covariates)} are not present in the clinical dataframe."
            )
        self.batch_column = batch_column
        if self.clinical_df[self.batch_column].nunique() <= 1:
            raise ValueError(
                "Cohort QC is not applicable to cohorts with only one batch."
            )
        # remove covariates with only one unique value
        self.covariates = [
            cov for cov in covariates if clinical_df[cov].unique().size > 1
        ]
        if len(self.covariates) < len(covariates):
            logging.warning(
                f"Since covariates {', '.join(np.setdiff1d(covariates, self.covariates))} have only one unique value, they are removed."
            )
        self.mixed_datasets = self.identify_mixed_datasets()

        self.data_expression_df = data_expression_df.loc[:, common_samples]
        # Check if data_expression_df_before is provided, if not, handle accordingly
        if data_expression_df_before is not None:
            self.data_expression_df_before = data_expression_df_before.loc[
                :, common_samples
            ]
        else:
            self.data_expression_df_before = None
        self.n_components = n_components
        self.n_neighbors = n_neighbors

    def identify_mixed_datasets(self) -> list:
        """
        Identify datasets that contain several combinations of covariates, indicating mixed datasets.

        Returns
        -------
        mixed_datasets: list
            A list of dataset names that contain multiple unique combinations of covariates.
        """

        # Group by dataset and covariates, then count unique combinations per dataset
        grouped = (
            self.clinical_df.groupby([self.batch_column] + self.covariates)
            .size()
            .reset_index(name="Counts")
        )

        # Count the number of unique combinations per dataset
        combination_counts = grouped.groupby(self.batch_column).size()

        # Identify datasets with more than one unique combination
        mixed_datasets = combination_counts[combination_counts > 1].index.tolist()

        return mixed_datasets

    def pca_plot(
        self,
        pcs: np.ndarray,
        labels: pd.Series,
        title: str = "PCA Plot",
        PC_list: list = [0, 1],
    ) -> plt.Figure:
        """
        Generates a PCA plot and returns it as a matplotlib figure object.

        Parameters
        ----------
        pcs: np.ndarray
            Array of principal components (e.g., from PCA).
        labels: pd.Series
            Series containing the labels to color the points by.
        title: str
            Title of the plot.

        Returns
        -------
        fig: plt.Figure
            The PCA plot as a figure object.
        """
        # Create the figure and axis
        fig, ax = plt.subplots(figsize=(12, 7))

        # Generate the scatter plot
        sns.scatterplot(
            x=pcs[:, PC_list[0]],
            y=pcs[:, PC_list[1]],
            hue=labels,
            palette="Set1",
            ax=ax,
        )

        # Set plot title and labels
        ax.set_title(title)
        ax.set_xlabel(f"PC{PC_list[0]}")
        ax.set_ylabel(f"PC{PC_list[1]}")

        # Position the legend outside the plot
        ax.legend(title=labels.name, bbox_to_anchor=(1.05, 1), loc="upper left")

        # Tight layout to adjust the plot
        plt.tight_layout()

        # Return the figure object
        return fig

    def plot_pca_variance(
        self, explained_variance: np.ndarray, ylim: float = None
    ) -> plt.Figure:
        """
        Generate a PCA variance explained barplot and return the figure object.

        Parameters
        ----------
        explained_variance: np.ndarray
            Array of variance explained ratios by each principal component.

        Returns
        -------
        fig: plt.Figure
            The PCA variance barplot as a figure object.
        """
        fig, ax = plt.subplots(figsize=(8, 6))
        ax.bar(range(1, len(explained_variance) + 1), explained_variance * 100)
        # Set y-axis limit to 10
        ax.set_ylim(0, ylim * 100)
        ax.set_xlabel("Principal Component")
        ax.set_ylabel("Variance Explained (%)")
        ax.set_title("PCA Variance Explained")

        # Tight layout to ensure everything fits within the figure
        plt.tight_layout()

        return fig

    def create_correlation_matrix_with_pc(
        self, clinical_data: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Create a correlation matrix between PC columns and other clinical data elements.

        - If elements are both numeric, we use Pearson correlation.
        - If one element is categorical and the other numeric, we use a t-test or ANOVA depending on the number of categories.

        Parameters
        ----------
        clinical_data: pd.DataFrame
            Clinical data containing PCs and other data elements.

        Returns
        -------
        pd.DataFrame
            A correlation matrix between PC columns and other clinical data elements.
        """

        # Filter PC columns (assuming continuous) and other data elements
        pc_columns = [col for col in clinical_data.columns if col.startswith("PC")]
        other_columns = [
            col for col in clinical_data.columns if not col.startswith("PC")
        ]

        # Initialize an empty correlation matrix
        correlation_matrix = pd.DataFrame(index=pc_columns, columns=other_columns).map(
            lambda x: (np.nan, np.nan, "Not computed", 0)
        )

        # Numerical data elements excluding PCs
        numerical_columns = np.intersect1d(
            clinical_data.select_dtypes(include=[float, int]).columns.tolist(),
            other_columns,
        )
        cat_columns = np.setdiff1d(other_columns, numerical_columns)

        # Pearson correlation between PC and other continuous numerical data
        for pc in pc_columns:
            for num_col in numerical_columns:
                if num_col != pc:
                    temp_data = clinical_data[[pc, num_col]].dropna()
                    if temp_data.shape[0] >= 2:
                        correlation_matrix.at[pc, num_col] = tuple(
                            [
                                round_scientific_notation(val)
                                for val in stats.pearsonr(
                                    temp_data[pc], temp_data[num_col]
                                )[:2]
                            ]
                            + ["Pearson correlation", temp_data.shape[0]]
                        )

        # T-test or ANOVA between PC and categorical data elements
        for pc in pc_columns:
            for cat_col in cat_columns:
                temp_data = clinical_data[[pc, cat_col]].dropna()
                if temp_data.empty:
                    continue
                categories = temp_data[cat_col].unique()
                if len(categories) == 2:
                    # T-test for binary categorical data
                    correlation_matrix.at[pc, cat_col] = tuple(
                        [
                            round_scientific_notation(val)
                            for val in stats.ttest_ind(
                                temp_data.loc[temp_data[cat_col] == categories[0], pc],
                                temp_data.loc[temp_data[cat_col] == categories[1], pc],
                            )[:2]
                        ]
                        + ["T-test", temp_data.shape[0]]
                    )
                elif len(categories) > 2:
                    # ANOVA for multi-category categorical data
                    try:
                        model = ols(f"{pc} ~ C(Q('{cat_col}'))", data=temp_data).fit()
                        anova_table = sm.stats.anova_lm(model, typ=2)
                        correlation_matrix.at[pc, cat_col] = tuple(
                            [
                                round_scientific_notation(val)
                                for val in anova_table.loc[
                                    f"C(Q('{cat_col}'))", ["F", "PR(>F)"]
                                ].values
                            ]
                            + ["ANOVA", temp_data.shape[0]]
                        )
                    except Exception as e:
                        if (
                            "r_matrix performs f_test for using dimensions that are asymptotically non-normal"
                            in str(e)
                        ):
                            correlation_matrix.at[pc, cat_col] = (
                                np.nan,
                                np.nan,
                                "Convergence error",
                                temp_data.shape[0],
                            )
                        else:
                            raise e

        # Mirror the matrix to fill both upper and lower triangle
        return correlation_matrix.T

    def _compute_pc_associations(self, pcs: np.ndarray) -> pd.DataFrame:
        """
        Compute the correlation between principal components and clinical annotations.

        Parameters
        ----------
        pcs: np.ndarray
            Array of principal components (e.g., from PCA).

        Returns
        -------
        correlation_matrix: pd.DataFrame
            A DataFrame containing the correlation between PCs and clinical annotations.
        """
        # Add all PCs to the clinical data at once
        pc_data = self.clinical_df.copy()
        for i in range(pcs.shape[1]):
            pc_data[f"PC{i + 1}"] = pcs[:, i]

        # Generate the correlation matrix once with all PCs
        correlation_matrix = self.create_correlation_matrix_with_pc(pc_data)
        return correlation_matrix

    def pca_analysis(self) -> tuple:
        """
        Perform PCA and compute associations between principal components and clinical annotations.

        Returns
        -------
        tuple
            Tuple containing the association matrices, PCA models, and principal components before and after correction.
        """
        # Perform PCA for data after batch effect correction
        pca_after = PCA(n_components=self.n_components)
        scaler = StandardScaler()
        pcs_after = pca_after.fit_transform(
            scaler.fit_transform(self.data_expression_df.T)
        )

        # If data_expression_df_before is not None, compute PCA for it
        if self.data_expression_df_before is not None:
            pca_before = PCA(n_components=self.n_components)
            pcs_before = pca_before.fit_transform(
                scaler.fit_transform(self.data_expression_df_before.T)
            )
            association_matrix_before = self._compute_pc_associations(pcs_before)
        else:
            pca_before = None
            pcs_before = None
            association_matrix_before = None

        # Compute association between PCs and clinical annotations for data after correction
        association_matrix = self._compute_pc_associations(pcs_after)

        return (
            association_matrix_before,
            association_matrix,
            pca_before,
            pca_after,
            pcs_before,
            pcs_after,
        )

    def cohort_summary(self) -> dict:
        """
        Generate a summary of the cohort, including the number of samples, datasets, genes, and covariate combinations.

        Returns
        -------
        dict
            Summary of the cohort with details on samples, datasets, genes, and covariate combinations.
        """
        # Generate a summary of the cohort
        num_samples = self.clinical_df.shape[0]
        num_datasets = self.clinical_df[self.batch_column].nunique()
        num_genes = self.data_expression_df.shape[0]
        covariates_used = ", ".join(self.covariates)

        # Calculate the number of samples by covariate combinations
        if len(self.covariates) > 0:
            covariate_combinations = (
                self.clinical_df.groupby(self.covariates)
                .size()
                .reset_index(name="Number of Samples")
            )
        else:
            covariate_combinations = pd.DataFrame(
                {"Number of Samples": [len(self.clinical_df)]}
            )

            # Convert the covariate combinations and their counts to a readable format
        covariate_combination_summary = covariate_combinations.to_dict(orient="records")

        summary = {
            "Number of Samples": num_samples,
            "Number of Datasets": num_datasets,
            "Number of Genes": num_genes,
            "Covariates Used": covariates_used,
            "Samples by Covariate Combination": covariate_combination_summary,
        }

        return summary

    def quantify_correction_effect(self) -> tuple:
        """
        Compute a metric quantifying the effect of batch correction on the variability of gene expression data.

        Returns
        -------
        tuple
            A tuple containing the mean median absolute deviation before and after correction, and the correction effect metric.
        """
        # Compute MAD for after correction
        mad_after = stats.median_abs_deviation(
            self.data_expression_df.to_numpy(), axis=1
        )

        if self.data_expression_df_before is not None:
            # Compute MAD for before correction if data is available
            mad_before = stats.median_abs_deviation(
                self.data_expression_df_before.to_numpy(), axis=1
            )
            effect_metric = np.mean(mad_after) / np.mean(mad_before)
            return mad_before.mean(), mad_after.mean(), effect_metric
        else:
            # Handle the case where there is no before-correction data
            return None, mad_after.mean(), None

    def compare_sample_distribution_by_covariates(
        self, combination: tuple
    ) -> plt.Figure:
        """
        Compare the distribution of the samples in the batch groups by a specific covariate combination.

        Parameters
        ----------
        combination : tuple
            A tuple of covariate values representing a unique covariate combination.

        Returns
        -------
        matplotlib.figure.Figure
            The generated plot object.
        """
        # Filter the clinical data to select samples corresponding to the current covariate combination
        filter_condition = self.clinical_df[self.covariates].eq(combination).all(axis=1)
        selected_samples = self.clinical_df[filter_condition].index

        data_after = self.data_expression_df[selected_samples]
        data_after_melted = data_after.melt(var_name="Sample", value_name="Expression")
        data_after_melted["Dataset"] = self.clinical_df.loc[
            data_after_melted["Sample"], self.batch_column
        ].values
        data_after_melted["Type"] = data_after_melted["Dataset"].apply(
            lambda x: "Mixed" if x in self.mixed_datasets else "Non-Mixed"
        )
        data_after_melted["Dataset"] = (
            data_after_melted["Dataset"].astype(str).apply(truncate_name, args=(11,))
        )

        # Plot the boxplot
        # Calculate the number of unique datasets
        num_datasets = len(data_after_melted["Dataset"].unique())

        # Dynamically adjust the width of the plot based on the number of datasets
        base_width = 3  # base width for a small number of datasets
        width_per_dataset = 0.65  # additional width per dataset
        fig_width = base_width + width_per_dataset * max(num_datasets - 1, 0)

        if self.data_expression_df_before is not None:
            fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(fig_width, 8))

            data_before = self.data_expression_df_before[selected_samples]
            data_before_melted = data_before.melt(
                var_name="Sample", value_name="Expression"
            )
            data_before_melted["Dataset"] = self.clinical_df.loc[
                data_before_melted["Sample"], self.batch_column
            ].values
            data_before_melted["Type"] = data_before_melted["Dataset"].apply(
                lambda x: "Mixed" if x in self.mixed_datasets else "Non-Mixed"
            )
            data_before_melted["Dataset"] = (
                data_before_melted["Dataset"]
                .astype(str)
                .apply(truncate_name, args=(11,))
            )
            # Plot before correction
            sns.violinplot(
                ax=axs[0],
                x="Dataset",
                y="Expression",
                hue="Type",
                data=data_before_melted,
                palette={"Mixed": "red", "Non-Mixed": "grey"},
            )
            axs[0].set_title("Gene expression distribution before correction")
            axs[0].set_xlabel("Dataset (Batch)")
            axs[0].set_ylabel("Gene Expression")
            axs[0].tick_params(axis="x", rotation=45)
            axs[0].legend(title="Dataset Type", loc="upper right")
            ax = axs[1]
        else:
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(fig_width, 5))

        # Plot after correction
        sns.violinplot(
            ax=ax,
            x="Dataset",
            y="Expression",
            hue="Type",
            data=data_after_melted,
            palette={"Mixed": "red", "Non-Mixed": "grey"},
        )
        ax.set_title("Gene expression distribution after correction")
        ax.set_xlabel("Dataset (Batch)")
        ax.set_ylabel("Gene Expression")
        ax.tick_params(axis="x", rotation=45)
        ax.legend(title="Dataset Type", loc="upper right")

        # Return the plot as a figure object
        fig.tight_layout()
        return fig

    def compare_sample_distribution_total(self) -> plt.Figure:
        """
        Compare the overall distribution of samples by dataset before and after correction across all covariate combinations.

        Returns
        -------
        matplotlib.figure.Figure
            The generated plot object.
        """
        self.clinical_df["Covariate_Combination"] = (
            self.clinical_df[self.covariates].astype(str).agg("-".join, axis=1)
        )

        data_after = self.data_expression_df.copy()
        mean_after = data_after.mean(axis=0).reset_index()
        mean_after.columns = ["Sample", "Mean_Expression"]
        mean_after["Covariate_Combination"] = self.clinical_df.loc[
            mean_after["Sample"], "Covariate_Combination"
        ].values

        width = (11 / (12 - min(11, mean_after["Covariate_Combination"].nunique()))) + 4
        height = max(
            10,
            (
                max(
                    len(item)
                    for item in mean_after["Covariate_Combination"].unique()
                    if isinstance(item, str)
                )
            )
            / 3,
        )

        if self.data_expression_df_before is not None:
            data_before = self.data_expression_df_before.copy()
            mean_before = data_before.mean(axis=0).reset_index()
            mean_before.columns = ["Sample", "Mean_Expression"]
            mean_before["Covariate_Combination"] = self.clinical_df.loc[
                mean_before["Sample"], "Covariate_Combination"
            ].values

            fig, axs = plt.subplots(nrows=2, ncols=1, figsize=(width, height))

            # Plot before correction
            sns.violinplot(
                ax=axs[0],
                x="Covariate_Combination",
                y="Mean_Expression",
                data=mean_before,
            )
            axs[0].set_title(
                "Mean of sample gene expression distribution before correction"
            )
            axs[0].set_xlabel("Covariate combination")
            axs[0].set_ylabel("Mean gene expression by sample")
            axs[0].tick_params(axis="x", rotation=45)
            ax = axs[1]
        else:
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(width, height / 2))

        # Plot after correction
        sns.violinplot(
            ax=ax,
            x="Covariate_Combination",
            y="Mean_Expression",
            data=mean_after,
        )
        ax.set_title("Mean of sample gene expression distribution after correction")
        ax.set_xlabel("Covariate combination")
        ax.set_ylabel("Mean gene expression by sample")
        ax.tick_params(axis="x", rotation=45)

        # Return the plot as a figure object
        fig.tight_layout()
        return fig

    def silhouette_score(self) -> tuple[float, float]:
        """
        Compute the Silhouette Score for batch mixing before and after batch correction.

        Returns
        -------
        tuple[float, float]
            Silhouette scores before and after correction.
        """
        # Compute silhouette score for after batch correction
        score_after = silhouette_score(
            self.data_expression_df.T, self.clinical_df[self.batch_column]
        )

        if self.data_expression_df_before is not None:
            # Compute silhouette score before correction if available
            score_before = silhouette_score(
                self.data_expression_df_before.T,
                self.clinical_df[self.batch_column],
            )
        else:
            score_before = None

        return score_before, score_after

    def compute_entropy(self, data_expression_df: pd.DataFrame) -> float:
        """
        Compute the entropy of batch mixing to evaluate the effectiveness of batch correction.

        Parameters
        ----------
        data_expression_df : pandas.DataFrame
            Gene expression data frame to evaluate.
        n_neighbors : int
            Number of nearest neighbors to consider in the entropy calculation.

        Returns
        -------
        float
            Mean entropy value indicating the level of batch mixing.
        """
        nbrs = NearestNeighbors(n_neighbors=self.n_neighbors, metric="euclidean").fit(
            data_expression_df.T
        )
        _, indices = nbrs.kneighbors(data_expression_df.T)
        batch_labels = self.clinical_df[self.batch_column].values

        entropies = []
        for index in indices:
            neighbors_labels = batch_labels[index]
            label_counts = np.unique(neighbors_labels, return_counts=True)[1]
            probabilities = label_counts / len(neighbors_labels)
            entropy = -np.sum(probabilities * np.log2(probabilities))
            entropies.append(entropy)

        return np.mean(entropies)

    def entropy_batch_mixing(self) -> tuple[float, float]:
        """
        Compute the entropy of batch mixing before and after batch correction.

        Returns
        -------
        tuple[float, float]
            Entropy values before and after correction.
        """
        # Compute entropy of batch mixing after batch correction
        entropy_after = self.compute_entropy(self.data_expression_df)

        if self.data_expression_df_before is not None:
            # Compute entropy of batch mixing before batch correction
            entropy_before = self.compute_entropy(self.data_expression_df_before)
        else:
            entropy_before = None

        return entropy_before, entropy_after

    def summarize_and_compare_mixed_datasets(self) -> dict:
        """
        Summarize and compare mixed datasets to non-mixed datasets to evaluate their representativeness.

        Returns
        -------
        dict
            Summary with details about each mixed dataset and all mixed datasets in general.
        """

        summary = {}

        # Identify mixed and non-mixed datasets
        non_mixed_datasets = np.setdiff1d(
            self.clinical_df[self.batch_column].unique(), self.mixed_datasets
        )

        summary["total_mixed_datasets"] = len(self.mixed_datasets)
        summary["total_non_mixed_datasets"] = len(non_mixed_datasets)

        # Overall summary for mixed and non-mixed datasets
        mixed_samples = self.clinical_df[
            self.clinical_df[self.batch_column].isin(self.mixed_datasets)
        ]
        non_mixed_samples = self.clinical_df[
            self.clinical_df[self.batch_column].isin(non_mixed_datasets)
        ]

        summary["total_mixed_samples"] = mixed_samples.shape[0]
        summary["total_non_mixed_samples"] = non_mixed_samples.shape[0]

        # Summarize by covariates (e.g., biopsy site)
        covariate_tuple = (
            self.covariates if isinstance(self.covariates, list) else [self.covariates]
        )

        if len(covariate_tuple) > 0:
            summary["mixed_samples_by_covariate"] = (
                mixed_samples.groupby(covariate_tuple).size().to_dict()
            )
            summary["non_mixed_samples_by_covariate"] = (
                non_mixed_samples.groupby(covariate_tuple).size().to_dict()
            )
        else:
            summary["mixed_samples_by_covariate"] = {(): len(mixed_samples)}
            summary["non_mixed_samples_by_covariate"] = {(): len(non_mixed_samples)}

        # Compare clinical annotations between mixed and non-mixed datasets
        comparison_results = {}
        for column in self.clinical_df.columns:
            if column not in [self.batch_column] + self.covariates:
                mixed_values = mixed_samples[column].value_counts(normalize=True)
                non_mixed_values = non_mixed_samples[column].value_counts(
                    normalize=True
                )
                comparison_results[column] = {
                    "mixed": mixed_values.to_dict(),
                    "non_mixed": non_mixed_values.to_dict(),
                }

        summary["clinical_annotation_comparison"] = comparison_results

        # Summary by each mixed dataset with composition and proportions
        mixed_dataset_details = {}
        for dataset in self.mixed_datasets:
            dataset_samples = self.clinical_df[
                self.clinical_df[self.batch_column] == dataset
            ]
            samples_by_combination = (
                dataset_samples.groupby(covariate_tuple).size().to_dict()
            )
            total_samples_by_combination = (
                self.clinical_df.groupby(covariate_tuple).size().to_dict()
            )

            proportion_by_combination = {
                comb: samples_by_combination.get(comb, 0)
                / total_samples_by_combination[comb]
                for comb in total_samples_by_combination
            }

            details = {
                "total_samples": dataset_samples.shape[0],
                "samples_by_covariate_combination": samples_by_combination,
                "proportion_by_covariate_combination": proportion_by_combination,
            }
            mixed_dataset_details[dataset] = details

        summary["mixed_dataset_details"] = mixed_dataset_details

        # Compute the overall proportion of mixed samples by covariate combination across all datasets
        if len(covariate_tuple) > 0:
            total_samples_by_combination = (
                self.clinical_df.groupby(covariate_tuple).size().to_dict()
            )
            mixed_samples_by_combination = (
                mixed_samples.groupby(covariate_tuple).size().to_dict()
            )
        else:
            total_samples_by_combination = {(): len(self.clinical_df)}
            mixed_samples_by_combination = {(): len(mixed_samples)}

        overall_proportion_by_combination = {
            comb: mixed_samples_by_combination.get(comb, 0)
            / total_samples_by_combination[comb]
            for comb in total_samples_by_combination
        }

        summary["overall_proportion_by_covariate_combination"] = (
            overall_proportion_by_combination
        )

        return summary

    def process(self) -> None:
        """
        Run all QC checks including PCA analysis, batch correction assessment.
        """

        # Run all QC checks
        (
            self.association_matrix_before,
            self.association_matrix,
            self.pca_before,
            self.pca_after,
            self.pcs_before,
            self.pcs_after,
        ) = self.pca_analysis()
        self.mad_before, self.mad_after, self.effect_metric = (
            self.quantify_correction_effect()
        )
        self.silhouette_before, self.silhouette_after = self.silhouette_score()
        self.entropy_before, self.entropy_after = self.entropy_batch_mixing()
        self.summary = self.summarize_and_compare_mixed_datasets()



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# original code was contributed to StackOverflow: https://stackoverflow.com/questions/71550468/does-python-have-an-analogue-to-rs-splinesns
# code was improved to better match the R code

# As per StackOverflow terms of service (see https://stackoverflow.com/help/licensing), original code is licensed under CC BY SA 4.0, which is compatible with GPL3 (see https://creativecommons.org/2015/10/08/cc-by-sa-4-0-now-one-way-compatible-with-gplv3/)

import numpy as np
from scipy.interpolate import splev

from ..utils import LOGGER


def spline_design(knots, x, order, derivs=0):
    """
    Evaluate the design matrix for the B-splines defined by :code:`knots` at the values in :code:`x`.

    Arguments
    ---------
    knots : array_like
        vector of knot positions (which will be sorted increasingly if needed).
    x : array_like
        vector of values at which to evaluate the B-spline functions or
        derivatives. The values in x must be between the “inner” knots
        :code:`knots[ord]` and :code:`knots[ length(knots) - (ord-1)]`
    order : int
        a positive integer giving the order of the spline function. This is the
        number of coefficients in each piecewise polynomial segment, thus a
        cubic spline has order 4.
    derivs : array_like
        an integer vector with values between 0 and :code:`ord` - 1,
        conceptually recycled to the length of :code:`x`. The derivative of the
        given order is evaluated at the :code:`x` positions. Defaults to zero.

    Returns
    -------
    ndarray
        a matrix with :code:`len(x)` rows and :code:`len(knots) - ord` columns.
        The :code:`i`'th row of the matrix contains the coefficients of the
        B-splines (or the indicated derivative of the B-splines) defined by the
        knot vector and evaluated at the :code:`i`'th value of :code:`x`. Each
        B-spline is defined by a set of :code:`ord` successive knots so the
        total number of B-splines is :code:`len(knots) - ord`.
    """
    derivs = np.asarray(derivs)
    if derivs.ndim == 0:
        der = np.repeat(derivs, len(x))
    else:
        der = np.zeros(len(x), dtype=int)
        der[: len(derivs)] = derivs
    n_bases = len(knots) - order
    res = np.empty((len(x), n_bases), dtype=float)
    for i in range(n_bases):
        coefs = np.zeros((n_bases,))
        coefs[i] = 1
        for j in range(len(x)):
            res[j, i] = splev(x, (knots, coefs, order - 1), der=der[j])[j]
    return res


class ns:
    """
    Class storing the B-spline basis matrix for a natural cubic spline and info used to generate it.

    Attributes
    ----------
    knots : array_like
        breakpoints that define the spline.
    include_intercept : bool
        whether an intercept is included in the basis
    boundary_knots : array_like
        boundary points at which to impose the natural boundary conditions and
        anchor the B-spline basis.
    basis : ndarray
        a matrix of dimension :code:`(len(x), df)`, where :code:`df =
        len(knots)-1-intercept` if :code:`df` was not supplied.
    """

    def __init__(
        self, x, df=None, knots=None, boundary_knots=None, include_intercept=False
    ):
        """
        Generate the B-spline basis matrix for a natural cubic spline.

        This function intends to provide the same functionality as R splines::ns.

        Arguments
        ---------
        x : array_like
            the predictor variable
        df : int, optional
            degrees of freedom. If :code:`knots` is not specified, then the
            function chooses :code:`df - 1 - intercept` knots at suitably chosen
            quantiles of :code:`x`. If :code:`None`, the number of inner knots is
            set to :code:`len(knots)`.
        knots : array_like, optional
            breakpoints that define the spline. The default is no knots; together
            with the natural boundary conditions this results in a basis for linear
            regression on :code:`x`. Typical values are the mean or median for one
            knot, quantiles for more knots. See also :code:`boundary_knots`.
        include_intercept : bool, optional
            if :code:`True`, an intercept is included in the basis; default is
            :code:`False`.
        boundary_knots : array_like, optional
            boundary points at which to impose the natural boundary conditions and
            anchor the B-spline basis (default the range of the data). If both
            :code:`knots` and :code:`boundary_knots` are supplied, the basis
            parameters do not depend on :code:`x`. Data can extend beyond
            :code:`boundary_knots`.

        Returns
        -------
        ndarray
            a matrix of dimension :code:`(len(x), df)`, where :code:`df =
            len(knots)-1-intercept` if :code:`df` was not supplied.
        """
        self.include_intercept = include_intercept
        x = np.asarray(x)
        if boundary_knots is None:
            boundary_knots = [np.min(x), np.max(x)]
            outside = False
        else:
            boundary_knots = list(np.sort(boundary_knots))
            out_left = x < boundary_knots[0]
            out_right = x > boundary_knots[1]
            outside = out_left | out_right
        self.boundary_knots = boundary_knots

        if df is not None and knots is None:
            nIknots = df - 1 - include_intercept
            if nIknots < 0:
                nIknots = 0
                LOGGER.warning("df was too small, used {1+include_intercept}")

            if nIknots > 0:
                knots = np.linspace(0, 1, num=nIknots + 2)[1:-1]
                knots = np.quantile(x, knots)
        else:
            nIknots = len(knots)
        self.knots = knots

        Aknots = np.sort(np.concatenate((boundary_knots * 4, knots)))

        if np.any(outside):
            basis = np.empty((x.shape[0], nIknots + 4), dtype=float)
            if np.any(out_left):
                k_pivot = boundary_knots[0]
                xl = np.ones((np.sum(out_left), 2))
                xl[:, 1] = x[out_left] - k_pivot
                tt = spline_design(Aknots, [k_pivot, k_pivot], 4, [0, 1])
                basis[out_left, :] = xl @ tt
            if np.any(out_right):
                k_pivot = boundary_knots[1]
                xr = np.ones((np.sum(out_right), 2))
                xr[:, 1] = x[out_right] - k_pivot
                tt = spline_design(Aknots, [k_pivot, k_pivot], 4, [0, 1])
                basis[out_right, :] = xr @ tt
            inside = ~outside
            if np.any(inside):
                basis[inside, :] = spline_design(Aknots, x[inside], 4)
        else:
            basis = spline_design(Aknots, x, 4)

        const = spline_design(Aknots, boundary_knots, 4, [2, 2])

        if include_intercept is False:
            basis = basis[:, 1:]
            const = const[:, 1:]

        qr_const = np.linalg.qr(const.T, mode="complete")[0]
        self.basis = (qr_const.T @ basis.T).T[:, 2:]

    def predict(self, newx):
        """
        Evaluate the spline basis at given values

        Arguments
        ---------
        newx : ndarray
            new predictor variable to regenerate the spline from

        Returns
        -------
        ns
            a new natural spline object, evaluated at the given values
        """
        return ns(
            newx,
            knots=self.knots,
            boundary_knots=self.boundary_knots,
            include_intercept=self.include_intercept,
        )



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is a Python port of the original C++ code from the file
# 'src/R_initialize_levenberg.cpp' of the Bioconductor edgeR package (version
# 3.38.4).

import numpy as np
from scipy.linalg import qr, solve_triangular


def get_levenberg_start(y, offset, disp, weights, design, use_null):
    """
    Compute an initial value for the parameters of the NB GLM

    Arguments
    ---------
    y : array_like
        matrix of counts
    offset : array_like
        matrix of offsets, same shape as :code:`y`
    disp : array_like
        matrix of dispersions, broadcastable to the shape of :code:`y`
    weights : array_like
        matrix of observation weights
    design : array_like
        design matrix
    use_null : bool
        whether to use the null method

    Returns
    -------
    ndarray
        vector of initial values for beta
    """

    if weights is None:
        weights = 1.0

    M, N = design.shape
    K = np.minimum(M, N)
    res = np.zeros((y.shape[0], N))

    if use_null:
        Q, R = qr(design, mode="economic")
        assert Q.shape == (M, K)
        assert R.shape == (K, N)

        # computing weighted average of the count:library size ratios
        curN = np.exp(offset)
        curweight = weights * curN / (1 + disp * curN)
        curweight = np.broadcast_to(curweight, y.shape)
        sum_exprs = np.sum(y * curweight / curN, axis=1)
        sum_weight = np.sum(curweight, axis=1)
        with np.errstate(divide="ignore"):
            current = np.log(sum_exprs / sum_weight)

        with np.errstate(invalid="ignore"):
            for tag in range(y.shape[0]):
                # performing the QR decomposition and taking the solution
                res[tag] = solve_triangular(
                    R[:, :K],
                    Q.T @ np.repeat(current[tag], y.shape[1]),
                    check_finite=False,
                )

    else:
        # finding the delta
        delta = np.min([np.max(y), 1.0 / 6])
        # computing normalized log-expression values
        current = np.log(np.maximum(delta, y)) - offset
        current *= np.sqrt(weights)

        for tag in range(y.shape[0]):
            Q, R = qr(np.sqrt(weights[tag, :]) * design, mode="economic")
            assert Q.shape == (M, K)
            assert R.shape == (K, N)

            # performing the QR decomposition and taking the solution
            res[tag] = solve_triangular(
                R[:, :K], Q.T @ current[tag, :], check_finite=False
            )

    return res



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024-2025 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/toptable.R' of the Bioconductor limma package (version 3.55.1).

import logging

import numpy as np
import pandas as pd
import scipy
from statsmodels.stats.multitest import multipletests

from ..diffexp import DEResults
from .marraylm import MArrayLM


def topTable(
    fit,
    coef=None,
    number=10,
    genelist=None,
    adjust_method="fdr_bh",
    sort_by="B",
    resort_by=None,
    p_value=1,
    fc=None,
    lfc=None,
    confint=False,
):
    """
    Extract a table of the top-ranked genes from a linear model fit

    This function summarizes the linear model fit object produced by
    :func:`lmFit`, :func:lm_series`, :func:`mrlm` by selecting the top-ranked
    genes for any given contrast, or for a set of contrasts. It assumes that
    the linear model fit has already been processed by :func:`eBayes`.

    If :code:`coef` has a single value, then the moderated *t*-statistics and
    *p*-values for that coefficient or contrast are used. If :code:`coef` takes
    two or more values, the moderated *F*-statistics for that set of
    coefficients or contrasts are used. If :code:`coef=None`, then all the
    coefficients or contrasts in the fitted model are used, except that any
    coefficient named :code:`Intercept` will be removed.

    The *p*-values for the coefficient/contrast of interest are adjusted for
    multiple testing by a call to
    :func:`statsmodels.stats.multitest.multipletests`. The :code:`"fdr_bh"`
    method, which controls the expected false discovery rate (FDR) below the
    specified value, is the default adjustment method because it is the most
    likely to be appropriate for microarray studies. Note that the adjusted
    *p*-values from this method are bounds on the FDR rather than *p*-values in
    the usual sense. Because they relate to FDRs rather than rejection
    probabilities, they are sometimes called *q*-values.

    Note, if there is no good evidence for differential expression in the
    experiment, that it is quite possible for all the adjusted *p*-values to be
    large, even for all of them to be equal to one. It is quite possible for
    all the adjusted *p*-values to be equal to one if the smallest *p*-values
    is no smaller than :code:`1/ngenes` where :code:`ngenes` is the number of
    genes with non-missing *p*-values.

    The :code:`sort_by` argument specifies the criterion used to select the top
    genes. The choices are :code:`"logFC"` to sort by the (absolute)
    coefficient representing the log-fold-change; :code:`"A"` to sort by
    average expression level (over all arrays) in descending order, :code:`"T"`
    or :code:`"t"` for absolute *t*-statistic; :code:`"P"` or :code:`"p"` for
    *p*-values; or :code:`"B"` for the :code:`lods` or *B*-statistic.

    Normally the genes appear in order of selection in the output table. If a
    different order is wanted, then the :code:`resort_by` argument may be
    useful. For example, :code:`topTable(fit, sort_by="B", resort_by="logFC")`
    selects the top genes according to log-odds of differential expression and
    then orders the selected genes by log-ratio in decreasing order. Or
    :code:`topTable(fit, sort_by="logFC", resort_y="logFC")` would select the
    genes by absolute log-fold-change and then sort them from most positive to
    most negative.

    Toptable output for all probes in original (unsorted) order can be obtained
    by :code:`topTable(fit, sort="none", number=np.inf)`.

    By default :code:`number` probes are listed. Alternatively, by specifying
    :code:`p_values` and :code:`number=np.inf`, all genes with adjusted
    *p*-values below a specified value can be listed.

    The arguments :code:`fc` and :code:`lfc` give the ability to filter genes
    by log-fold change, but see the Notes below.

    Notes
    -----
    Although this function enables users to set both *p*-values and fold-change
    cutoffs, the use of fold-change cutoffs is not generally recommended. If
    the fold changes and *p*-values are not highly correlated, then the use of
    a fold change cutoff can increase the false discovery rate above the
    nominal level. Users wanting to use fold change thresholding are usually
    recommended to use :func:`treat` and :func:`topTreat` instead.

    In general, the adjusted *p*-values returned by
    :code:`adjust_method="fdr_bh"` remain valid as FDR bounds only when the
    genes remain sorted by *p*-value. Resorting the table by log-fold-change
    can increase the false discovery rate above the nominal level for genes at
    the top of resorted table.

    Arguments
    ---------
    fit : MArrayLM
        linear model fit produced by :func:`lmFit`, :func:`lm_series`,
        :func:`gls_series` or :func:`mrlm`
    coef : int or str
        column number or column name specifying which coefficient or contrast
        of the linear model is of interest. Can also be a vector of column
        subscripts, in which case the gene ranking is by *F*-statistic for that
        set of contrasts.
    number : int
        maximum number of genes to list
    genelist : array_like
        data frame or array containing gene information. Defaults to :code:`fit.genes`
    adjust_method : {"none", "fdr_bh", "fdr_by", "holm"}
        method used to adjust the *p*-values for multiple testing. See
        :func:`statsmodels.stats.multitest.multipletests` for the complete list
        of options. A :code:`None` value will result in the default adjustment
        method, which is :code:`"fdr_bh"`.
    sort_by : { "logFC", "log2FoldChange", "AveExpr", "t", "P", "p", "B", "none" }
        string specifying which statistic to rank the genes by
    resort_by : { "logFC", "AveExpr", "t", "P", "p", "B", "none" }
        string specifying statistic to sort the selected genes by in the output
        data frame
    p_value : float
        cutoff value for adjusted *p*-values. Only genes with lower *p*-values
        are listed
    fc : float, optional
        minimum fold-change required
    lfc : float, optional
        optional minimum log2-fold-change required, equal to :code:`log2(fc)`.
        :code:`fc` and :code:`lfc` are alternative ways to specify a
        fold-change cutoff and, if both are specified, then :code:`fc` takes
        precedence. If specified, then the results will include only genes with
        (at least one) absolute log-fold-change greater than :code:`lfc`
    confint : bool or float
        whether the confidence 95% intervals should be output for
        :code:`logFC`. Alternatively, can be a value between 0 and 1 specifying
        the required confidence level.

    Returns
    -------
    DEResults or pd.DataFrame
        :class:`~DEResults` if :code:`coef` has a single value, otherwise
        :class:`pd.DataFrame`. A dataframe with a row for each of the
        :code:`number` top genes and the following columns:

        - genelist: one or more columns of probe annotation, if
          :code:`genelist` was included as input
        - log2FoldChange: estimate of the log2-fold-change corresponding to the
          effect or contrast (:class:`DEResults` only)
        - CI_L: left limit of confidence interval for :code:`logFC`, if
          :code:`confint=True` or :code:`confint` is a numeric value
        - CI_R: right limit of confidence interval for :code:`logFC`, if
          :code:`confint=True` or :code:`confint` is a numeric value
        - AveExpr: average log2-expression for the probe over all arrays and
          channels, same as :code:`Amean` in the :class:`MArrayLM` object
        - stat: moderated *t*-statistic (:class:`DEResults` only)
        - F: moderated *F*-statistic (:class:`pd.DataFrame` only)
        - pvalue: raw *p*-value
        - adj_pvalue: adjusted *p*-value or *q*-value
        - B: log-odds that the gene is differentially expressed
    """
    if not isinstance(fit, MArrayLM):
        raise ValueError("fit must be a MArrayLM object")
    if (fit.t is None) and (fit.F is None):
        raise ValueError("Need to run eBayes or treat first")
    if fit.coefficients is None:
        raise ValueError("coefficients not found in fit object")
    if confint and (fit.stdev_unscaled is None):
        raise ValueError("stdev_unscaled not found in fit object")

    if genelist is None:
        genelist = fit.genes

    # Check coef
    if coef is None:
        if (not hasattr(fit, "treat_lfc")) or (fit.treat_lfc is None):
            coef = np.arange(fit.coefficients.shape[1])
        else:
            coef = fit.coefficients.shape[1]

    # Set log2-fold-change cutoff
    if fc is None:
        if lfc is None:
            lfc = 0
    else:
        if fc < 1:
            raise ValueError("fc must be greater than or equal to 1")
        lfc = np.log2(fc)

    # If testing for multiple coefficients, call low-level topTable function for F-statistics
    if not isinstance(coef, (int, str)):
        if len(coef) == 1:
            coef = coef[0]
    if not isinstance(coef, (int, str)):
        if hasattr(fit, "treat_lfc") and (fit.treat_lfc is not None):
            raise ValueError(
                "Treat p-values can only be displayed for single coefficient"
            )
        coef = np.unique(coef)
        if len(coef) < fit.coefficients.shape[1]:
            fit = fit[:, coef]
        if sort_by == "B":
            sort_by = "F"
        return _topTableF(
            fit,
            number=number,
            genelist=genelist,
            adjust_method=adjust_method,
            sort_by=sort_by,
            p_value=p_value,
            lfc=lfc,
        )

    # Call low-level topTabel function for t-statistics
    return _topTableT(
        fit,
        coef=coef,
        number=number,
        genelist=genelist,
        A=fit.Amean,
        eb=fit,
        adjust_method=adjust_method,
        sort_by=sort_by,
        resort_by=resort_by,
        p_value=p_value,
        lfc=lfc,
        confint=confint,
    )


def _topTableF(
    fit, number=10, genelist=None, adjust_method="fdr_bh", sort_by="F", p_value=1, lfc=0
):
    if genelist is None:
        genelist = fit.genes
    # Check fit
    if fit.coefficients is None:
        raise ValueError("coefficients not found in fit")
    M = fit.coefficients
    Amean = fit.Amean
    Fstat = fit.F
    Fp = fit.F_p_value
    if Fstat is None:
        raise ValueError("F-statistics not found in fit")

    # Ensure genelist is a data frame
    if (genelist is not None) and not isinstance(genelist, pd.DataFrame):
        genelist = pd.DataFrame({"ProbeID": genelist})

    # Check row names
    rn = M.index
    if len(np.unique(rn)) != len(rn):
        if genelist is None:
            genelist = pd.DataFrame({"ID": rn})
        else:
            if "ID" in genelist.columns:
                genelist["ID0"] = rn
            else:
                genelist["ID"] = rn

    # Check sort_by
    if sort_by not in ["F", "none"]:
        raise ValueError(f"invalid value {sort_by} for argument sort_by")

    # Apply multiple testing adjustment
    adj_pvalue = multipletests(Fp, method=adjust_method)[1]

    # Thin out fit by lfc and p_value thresholds
    if lfc > 0 or p_value < 1:
        if lfc > 0:
            big = np.nansum(np.abs(M) > lfc, axis=1) > 0
        else:
            big = True
        if p_value < 1:
            sig = adj_pvalue <= p_value
            sig[np.isnan(sig)] = False
        else:
            sig = True
        keep = big & sig
        if not np.all(keep):
            M = M[keep, :]
            rn = rn[keep]
            Amean = Amean[keep]
            Fstat = Fstat[keep]
            Fp = Fp[keep]
            genelist = genelist[keep, :]
            adj_pvalue = adj_pvalue[keep]

    # Enough rows left?
    if M.shape[0] < number:
        number = M.shape[0]
    if number < 1:
        return pd.DataFrame()

    # Find rows of top genes
    if sort_by == "F":
        o = np.argsort(Fp)[:number]
    else:
        o = np.arange(number)

    # Assemble data frame
    if genelist is None:
        tab = pd.DataFrame(M.iloc[o, :])
    else:
        tab = pd.DataFrame(genelist[o, :], M[o, :])
    tab["AveExpr"] = Amean.iloc[o]
    tab["F"] = Fstat[o]
    tab["pvalue"] = Fp[o]
    tab["adj_pvalue"] = adj_pvalue[o]
    tab.index = rn[o]
    return tab


def _topTableT(
    fit,
    coef=1,
    number=10,
    genelist=None,
    A=None,
    eb=None,
    adjust_method="fdr_bh",
    sort_by="B",
    resort_by=None,
    p_value=1,
    lfc=0,
    confint=False,
):
    # Check fit
    rn = fit.coefficients.index

    if not isinstance(coef, (int, str)):
        coef = coef[0]
        logging.warnings.warn(
            "Treat is for single coefficients: only first value of coef being used"
        )

    # Ensure genelist is a data frame
    if (genelist is not None) and not isinstance(genelist, pd.DataFrame):
        genelist = pd.DataFrame({"ID": genelist})

    # Check rownames
    if len(np.unique(rn)) != len(rn):
        if genelist is None:
            genelist = pd.DataFrame({"ID": rn})
        else:
            if "ID" in genelist.columns:
                genelist["ID0"] = rn
            else:
                genelist["ID"] = rn
        rn = np.arange(fit.coefficients.shape[0])

    # Check sort_by
    if sort_by not in [
        "logFC",
        "log2FoldChange",
        "M",
        "A",
        "Amean",
        "AveExpr",
        "P",
        "p",
        "T",
        "t",
        "B",
    ]:
        raise ValueError(f"invalid value {sort_by} for sort_by")
    if sort_by == "M" or sort_by == "log2FoldChange":
        sort_by = "logFC"
    if sort_by == "A" or sort_by == "Amean":
        sort_by = "AveExpr"
    if sort_by == "T":
        sort_by = "t"
    if sort_by == "p":
        sort_by = "P"

    # Check resort_by
    if resort_by is not None:
        if resort_by not in [
            "logFC",
            "log2FoldChange",
            "M",
            "A",
            "Amean",
            "AveExpr",
            "P",
            "p",
            "T",
            "t",
            "B",
        ]:
            raise ValueError(f"invalid value {sort_by} for sort_by")
        if resort_by == "M" or resort_by == "log2FoldChange":
            resort_by = "logFC"
        if resort_by == "A" or resort_by == "Amean":
            resort_by = "AveExpr"
        if resort_by == "T":
            resort_by = "t"
        if resort_by == "p":
            resort_by = "P"

    # Check A
    if A is None:
        if sort_by == "A":
            raise ValueError("Cannot sort by A-values as these have not been given")
    else:
        if A.ndim == 2:
            A = np.nanmean(A, axis=1)

    # Check for lods component
    if eb.lods is None:
        if sort_by == "B":
            raise ValueError(
                "Trying to sort_by B, but B-statistic (lods) not found in MArrayLM object"
            )
        if resort_by == "B":
            raise ValueError(
                "Trying to resort_by B, but B-statistic (lods) not found in MArrayLM object"
            )
        include_B = False
    else:
        include_B = True

    # Extract statistics for table
    M = fit.coefficients.loc[:, coef]
    lfcSE = fit.stdev_unscaled.loc[:, coef]
    tstat = eb.t.loc[:, coef]
    P_Value = eb.p_value.loc[:, coef]
    if include_B:
        B = eb.lods.loc[:, coef]

    # Apply multiple testing adjustment
    adj_pvalue = multipletests(P_Value, method=adjust_method)[1]

    # Thin out fit by p_value and lfc thresholds
    if p_value < 1 or lfc > 0:
        sig = (adj_pvalue <= p_value) & (np.abs(M) >= lfc)
        if np.any(np.isnan(sig)):
            sig[np.isnan(sig)] = False
        if not np.any(sig):
            return pd.DataFrame()
        genelist = genelist[sig, :]
        M = M[sig]
        lfcSE = lfcSE[sig]
        A = A[sig]
        tstat = tstat[sig]
        P_Value = P_Value[sig]
        adj_pvalue = adj_pvalue[sig]
        if include_B:
            B = B[sig]
        rn = rn[sig]

    # Are enough rows left?
    if len(M) < number:
        number = len(M)
    if number < 1:
        return pd.DataFrame()

    # Select top rows
    if sort_by == "logFC":
        ord = np.flip(np.argsort(np.abs(M)))
    elif sort_by == "AveExpr":
        ord = np.flip(np.argsort(A))
    elif sort_by == "P":
        ord = np.argsort(P_Value)
    elif sort_by == "t":
        ord = np.flip(np.argsort(np.abs(tstat)))
    elif sort_by == "B":
        ord = np.flip(np.argsort(B))
    elif sort_by == "none":
        ord = np.arange(len(M))
    top = ord[:number]

    # Assemble output data frame
    if genelist is None:
        tab = pd.DataFrame({"log2FoldChange": M.iloc[top], "lfcSE": lfcSE.iloc[top]})
    else:
        tab = pd.DataFrame(genelist.loc[top, :])
        tab["log2FoldChange"] = M[top]
        tab["lfcSE"] = lfcSE[top]

    if confint is not False:
        if isinstance(confint, (int, float)):
            alpha = (1 + confint) / 2
        else:
            alpha = 0.975
        margin_error = (
            np.sqrt(eb.s2_post.iloc[top])
            * fit.stdev_unscaled[coef].iloc[top]
            * scipy.stats.t.ppf(alpha, df=eb.df_total[top])
        )
        tab["CI_L"] = M.iloc[top] - margin_error
        tab["CI_R"] = M.iloc[top] + margin_error

    if A is not None:
        tab["AveExpr"] = A.iloc[top]
    tab["stat"] = tstat.iloc[top]
    tab["pvalue"] = P_Value.iloc[top]
    tab["adj_pvalue"] = adj_pvalue[top]

    if include_B:
        tab["B"] = B.iloc[top]
    tab.index = rn[top]

    # Resort table
    if resort_by is not None:
        if resort_by == "logFC":
            ord = np.flip(np.argsort(tab["log2FoldChange"]))
        elif resort_by == "AveExpr":
            ord = np.flip(np.argsort(tab["AveExpr"]))
        elif resort_by == "P":
            ord = np.argsort(tab["pvalue"])
        elif resort_by == "t":
            ord = np.flip(np.argsort(tab["stat"]))
        elif resort_by == "B":
            ord = np.flip(np.argsort(tab["B"]))
        tab = tab[ord, :]

    return DEResults(tab)



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/fitFDist.R' of the Bioconductor limma package (version 3.55.1).

import numpy as np
from scipy.special import digamma, polygamma

from ..utils import LOGGER, lm_fit, ns


def fitFDist(x, df1, covariate=None):
    """
    Moment estimation of the parameters of a scaled F-distribution given one of the degrees of freedom.

    This function is called internally by :func:`eBayes` and :func:`squeezeVar`
    and is not usually called directly by a user.

    This function implements an algorithm proposed by [Smyth2004]_ and
    [Phipson2016]_. It estimates :code:`scale` and :code:`df2` under the
    assumption that :code:`x` is distributed as :code:`scale` times an
    F-distributed random variable on :code:`df1` and :code:`df2` degrees of
    freedom. The parameters are estimated using the method of moments,
    specifically from the mean and variance of the :code:`x` values on the
    log-scale.

    When :code:`covariate` is supplied, a spline curve trend will be estimated
    for the :code:`x` values and the estimation will be adjusted for this trend
    [Phipson2016]_.

    Arguments
    ---------
    x : array_like
        1-D array of positive values representing a sample from a scaled F-distribution
    df1 : array_like
        the first degrees of freedom of the F-distribution. Can be a single
        value or an array of the same length as :code:`x`.
    covariate :
        if not :code:`None`, the estimated scale value will depend on this
        numeric covariate.

    Returns
    -------
    dict
        a dictionary with the following components:

        - :code:`"scale"`, a scale factor for F-distribution. An array if
          :code:`covariate` is not :code:`None`, a scalar otherwise.
        - :code:`"df2"`, the second degrees of freedom of the fitted F-distribution.
    """

    # Check x
    n = len(x)
    if n == 0:
        return {"scale": np.nan, "df2": np.nan}
    if n == 1:
        return {"scale": x, "df2": 0}

    # Check df1
    df1 = np.asarray(df1)
    ok = np.isfinite(df1) & (df1 > 1e-15)
    if df1.ndim == 0:
        if not ok:
            return {"scale": np.nan, "df2": np.nan}
        else:
            ok = np.full(n, True)
    else:
        if len(df1) != n:
            raise ValueError("x and df1 have different lengths")
    df1 = np.broadcast_to(df1, x.shape)

    # Check covariate
    if covariate is None:
        splinedf = 1
    else:
        covariate = np.asarray(covariate)
        if len(covariate) != n:
            raise ValueError("x and covariate must be of same length")
        if np.isnan(covariate).any():
            raise ValueError("NA covariate values not allowed")
        isfin = np.isfinite(covariate)
        if not isfin.all():
            if isfin.any():
                covariate[np.isneginf(covariate)] = np.min(covariate[isfin]) - 1
                covariate[np.isposinf(covariate)] = np.max(covariate[isfin]) + 1
            else:
                covariate = np.sign(covariate)

    # Remove missing or infinite or negative values and zero degrees of freedom
    ok = ok & np.isfinite(x) & (x > -1e-15)
    nok = ok.sum()
    if nok == 1:
        return {"scale": x[ok], "df2": 0}
    notallok = nok < n
    if notallok:
        x = x[ok]
        df1 = df1[ok]
        if covariate is not None:
            covariate_notok = covariate[~ok]
            covariate = covariate[ok]

    # Set df for spline trend
    if covariate is not None:
        splinedf = 1 + (nok >= 3) + (nok >= 6) + (nok >= 30)
        splinedf = min(splinedf, len(np.unique(covariate)))
        # If covariate takes only one unique value or insufficient observations, recall
        # with None covariate
        if splinedf < 2:
            out = fitFDist(x=x, df1=df1)
            out["scale"] = np.full(n, out["scale"])
            return out

    # Avoid exactly zero values
    x = np.maximum(x, 0)
    m = np.median(x)
    if m == 0:
        LOGGER.warning(
            "More than half of residual variances are exactly zero: eBayes unreliable"
        )
        m = 1
    else:
        if (x == 0).any():
            LOGGER.warning(
                "Zero sample variances detected, have been offset away from zero"
            )
    x = np.maximum(x, 1e-5 * m)

    # Better to work on with log(F)
    z = np.log(x)
    e = z - digamma(df1 / 2) + np.log(df1 / 2)

    if covariate is None:
        emean = e.mean()
        evar = np.sum((e - emean) ** 2 / (nok - 1))
    else:
        try:
            design = ns(covariate, df=splinedf, include_intercept=True)
        except:  # noqa: E722
            raise RuntimeError("Problem with covariate")
        fit = lm_fit(design.basis, e)
        if notallok:
            design2 = design.predict(newx=covariate_notok)
            emean = np.zeros(n)
            emean[ok] = fit.fitted_values
            emean[~ok] = design2.basis @ fit.coefficients
        else:
            emean = fit.fitted_values
        evar = (fit.effects[fit.rank :] ** 2).mean()

    # Estimate scale and df2
    evar = evar - polygamma(1, df1 / 2).mean()
    if evar > 0:
        df2 = 2 * trigammaInverse(evar)
        s20 = np.exp(emean + digamma(df2 / 2) - np.log(df2 / 2))
    else:
        df2 = np.inf
        if covariate is None:
            # Use simple pooled variance, which is MLE of the scale in this case.
            # Versions of limma before Jan 2017 returned the limiting value of the
            # evar>0 estimate, which is larger.
            s20 = x.mean()
        else:
            s20 = np.exp(emean)

    return {"scale": s20, "df2": df2}


def trigammaInverse(x):
    """
    Solve trigamma(y) = x for y

    Arguments
    ---------
    x : array_like

    Returns
    -------
    ndarray
        trigamma inverses for each value of :code:`x`
    """
    x = np.asarray(x)

    # Treat out-of-range values as special cases
    omit = np.isnan(x)
    if omit.any():
        y = x
        if (~omit).any():
            y[~omit] = trigammaInverse(x[~omit])
        return y

    omit = x < 0
    if omit.any():
        y = x
        y[omit] = np.nan
        LOGGER.warning("NaNs produced")
        if (~omit).any():
            y[~omit] = trigammaInverse(x[~omit])
        return y

    omit = x > 1e7
    if omit.any():
        y = x
        y[omit] = 1 / np.sqrt(x[omit])
        if (~omit).any():
            y[~omit] = trigammaInverse(x[~omit])
        return y

    omit = x < 1e-6
    if omit.any():
        y = x
        with np.errstate(divide="ignore"):
            y[omit] = 1 / x[omit]
        if (~omit).any():
            y[~omit] = trigammaInverse(x[~omit])
        return y

    # Newton's method
    # 1/trigamma(y) is convex, nearly linear and strictly > y-0.5,
    # so iteration to solve 1/x = 1/trigamma is monotonically convergent
    y = 0.5 + 1 / x
    it = 0
    while True:
        it += 1
        tri = polygamma(1, y)
        dif = tri * (1 - tri / x) / polygamma(2, y)
        y += dif
        if np.max(-dif / y) < 1e-8:
            break
        if it > 50:
            LOGGER.warning("Iteration limit exceeded")
            break
    return y



import unittest

import numpy as np

from inmoose.utils import cov2cor


class Test(unittest.TestCase):
    def test_cov2cor(self):
        M = np.arange(25).reshape((5, 5))
        cov = np.cov(M)
        cor = np.corrcoef(M)
        self.assertTrue(np.allclose(cov2cor(cov), cor))



import unittest

import numpy as np

from inmoose.limma import tmixture_matrix, tmixture_vector


class Test(unittest.TestCase):
    def test_tmixture_matrix(self):
        self.assertTrue(
            np.allclose(
                tmixture_matrix(
                    np.arange(1, 7).reshape(2, 3),
                    np.arange(2, 8).reshape(2, 3),
                    np.arange(3, 5),
                    0.5,
                ),
                np.array([648.387, 1544.773, 3105.003]),
            )
        )

    def test_tmixture_vector(self):
        self.assertAlmostEqual(
            tmixture_vector(
                np.arange(1, 7),
                np.arange(2, 8),
                np.arange(3, 9),
                0.5,
            ),
            1124.83,
            places=2,
        )



import unittest

import numpy as np

from inmoose.utils import lm_fit, lm_wfit


class Test(unittest.TestCase):
    def assertArrayEqual(self, x, y, tol=1e-6):
        self.assertTrue(np.allclose(x, y, atol=tol))

    def test_lm_fit(self):
        res = lm_fit(
            np.arange(5, 11).reshape(3, 2), np.array([1.0 / 30, 1.0 / 31, 1.0 / 32])
        )
        self.assertArrayEqual(res.coefficients, np.array([-0.03644713, 0.03592630]))
        self.assertArrayEqual(
            res.residuals, [1.120072e-05, -2.240143e-05, 1.120072e-05]
        )
        self.assertArrayEqual(res.fitted_values, [0.03332213, 0.03228047, 0.03123880])
        self.assertArrayEqual(res.effects, [-5.411474e-02, 1.413684e-02, 2.743604e-05])
        self.assertEqual(res.rank, 2)
        self.assertArrayEqual(res.df_residuals, [1])

    def test_lm_wfit(self):
        res = lm_wfit(
            np.arange(5, 11).reshape(3, 2),
            np.array([1.0 / 30, 1.0 / 31, 1.0 / 32]),
            np.array([100, 50, 1]),
        )
        self.assertArrayEqual(res.coefficients, np.array([-0.03654927, 0.03601318]))
        self.assertArrayEqual(
            res.residuals, [6.165532e-07, -2.466213e-06, 6.165532e-05]
        )
        self.assertArrayEqual(res.fitted_values, [0.03333272, 0.03226053, 0.03118834])
        self.assertArrayEqual(res.effects, [-3.981168e-01, 7.496570e-02, 6.437005e-05])
        self.assertEqual(res.rank, 2)
        self.assertArrayEqual(res.df_residuals, [1])



# -----------------------------------------------------------------------------
# Copyright (C) 2024 L. Meunier

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------


def round_scientific_notation(num: float) -> str:
    """Round a number in scientific notation to 2 decimals.

    Parameters
    ----------
    num: float
        number to round

    Returns
    -------
    str
        rounded number in scientific notation
    """
    return "{:0.2e}".format(num)


def truncate_name(name, max_length=10):
    """
    Truncates a given string to a maximum length.

    Parameters
    ----------
    name: str
        The input string to be truncated.
    max_length: int, optional
        The maximum allowed length of the truncated string. Defaults to 10.

    Returns
    -------
    str
        The truncated string, or the original string if it is already shorter than or equal to the maximum length.
    """
    return name if len(name) <= max_length else name[: (max_length - 3)] + "..."



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/exactTestBySmallP.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np

from ..utils import dnbinom_mu as dnbinom
from .binomTest import binomTest
from .exactTestDoubleTail import exactTestDoubleTail


def exactTestBySmallP(y1, y2, dispersion=0):
    """
    Compute genewise *p*-values for differences in the means between two groups of negative-binomially distributed counts.

    This function implements the method of small probabilities as proposed by
    [Robinson2008]_. This method corresponds exactly to :func:`binomTest` as the
    dispersion approaches zero, but gives poor results when the dispersion is
    very large.

    See also
    --------
    exactTest

    Arguments
    ---------
    y1 : matrix
        matrix of counts for the first of the two experimental groups to be
        tested for differences. Rows correspond to genes and columns to
        libraries. Libraries are assumed to be equal in size -- *e.g.* adjusted
        pseudocounts from the output of :func:`equalizeLibSizes`.
    y2 : matrix
        matrix of counts for the second of the two experimental groups to be
        tested for differences. Rows correspond to genes and columns to
        libraries. Libraries are assumed to be equal in size -- *e.g.* adjusted
        pseudocounts from the output of :func:`equalizeLibSizes`.
    dispersion : array_like of floats
        an array of dispersions, either of length one or of length equal to the
        number of genes.

    Returns
    -------
    ndarray
        array of genewise *p*-values, one for each row of :code:`y1` and :code:`y2`
    """
    y1 = np.asarray(y1)
    y2 = np.asarray(y2)
    if y1.shape[0] != y2.shape[0]:
        raise ValueError("Number of rows of y1 not equal to number of rows of y2")
    ntags = y1.shape[0]
    if np.isnan(y1).any() or np.isnan(y2).any():
        raise ValueError("NAs not allowed")
    n1 = y1.shape[1]
    n2 = y2.shape[1]

    if n1 == n2:
        return exactTestDoubleTail(y1=y1, y2=y2, dispersion=dispersion)

    dispersion = np.asarray(dispersion)
    sum1 = np.round(y1.sum(axis=1))
    sum2 = np.round(y2.sum(axis=1))
    if (dispersion == 0).all():
        return binomTest(sum1, sum2, p=n1 / (n1 + n2))
    if (dispersion == 0).any():
        raise ValueError("dispersion must be either all zero or all positive")
    dispersion = np.broadcast_to(dispersion, (ntags,))
    N = sum1 + sum2
    mu = N / (n1 + n2)
    r = 1 / dispersion
    all_zeros = N == 0

    pvals = np.ones(ntags)
    if ntags == 0:
        return pvals
    if any(all_zeros):
        pvals[~all_zeros] = exactTestBySmallP(
            y1=y1[~all_zeros, :],
            y2=y2[~all_zeros, :],
            dispersion=dispersion[~all_zeros],
        )
        return pvals

    for i in range(ntags):
        ind = np.arange(N[i] + 1)
        p_top = dnbinom(ind, size=n1 * r[i], mu=n1 * mu[i]) * dnbinom(
            N[i] - ind, size=n2 * r[i], mu=n2 * mu[i]
        )
        p_obs = dnbinom(sum1[i], size=n1 * r[i], mu=n1 * mu[i]) * dnbinom(
            sum2[i], size=n2 * r[i], mu=n2 * mu[i]
        )
        keep = p_top <= p_obs
        p_bot = dnbinom(N[i], size=(n1 + n2) * r[i], mu=(n1 + n2) * mu[i])
        pvals[i] = np.sum(p_top[keep] / p_bot)

    # edgeR code returns "min(pvals, 1)" but it looks like a typo: "pmin(pvals, 1)"
    # anyhow, we choose to align with edgeR behavior here
    # probably not a big deal as "smallp" is no longer recommended
    return np.min([np.min(pvals), 1])



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import copy
import logging
import re

import numpy as np
import pandas as pd
import patsy
import scipy

from ..utils import Factor, cov2cor


def makeContrasts(contrasts, levels):
    """
    Construct the contrast matrix corresponding to specified contrasts of a set of parameters.

    This function expresses contrasts between a set of parameters as a numeric
    matrix. The parameters are usually the coefficients from a linear model
    fit, so the matrix specifies which comparisons between the coefficients are
    to be extracted from the fit. The output from this function is usually used
    as input to :func:`contrasts_fit`.

    Arguments
    ---------
    contrasts : list
        strings parseable as expressions that describe the contrasts
        if :code:`contrasts` is a string, it will be interpreted as a list of
        one element
    levels : list or Factor or patsy.DesignMatrix
        list of strings or factor the names of the parameters of which
        contrasts are desired, or a design matrix or other object with the
        parameter names as column names.

    Returns
    -------
    pd.DataFrame
        matrix with columns corresponding to contrasts
    """
    if isinstance(levels, patsy.DesignMatrix):
        levels = levels.design_info.column_names
    levels = Factor(levels)

    def indicator(i, n):
        res = np.zeros(n)
        res[i] = 1
        return res

    def rename(n):
        return re.sub("([a-zA-Z0-9_]+)\[((T.)?[a-zA-Z0-9_]+)\]", "\\1_\\2", n)

    gl = {
        rename(L): indicator(i, levels.nlevels())
        for i, L in enumerate(levels.categories)
    }
    if not isinstance(contrasts, list):
        contrasts = [contrasts]
    return pd.DataFrame(
        {c: eval(rename(c), gl) for c in contrasts}, index=levels.categories
    )


def contrasts_fit(fit, contrasts=None, coefficients=None):
    """
    Compute contrasts from linear model fit

    Given a linear model fit to microarray data, compute estimated coefficients
    and standard errors for a given set of contrasts.

    This function accepts input from any of the functions :func:`lmFit`,
    :func:`lm_series`, :func:`mrlm`, :func:`gls_series` or :func:`lmscFit`. The
    function reorientates the fitted model object from the coefficients of the
    original design matrix to any set of contrasts of the original
    coefficients. The coefficients, unscaled standard deviations and
    correlation matrix are re-calculated in terms of the contrasts.

    The idea of this function is to fit a full-rank model using :func:`lmFit`
    or equivalent, then use :func:`contrasts_fit` to obtain coefficients and
    standard errors for any number of contrasts of the coefficients of the
    original model.  Unlike the design matrix input to :func:`lmFit`, which
    normally has one column for each treatment in the experiment, the matrix
    :code:`contrasts` may have any number of columns and these are not required
    to be linearly independent. Methods of assessing differential expression,
    such as :func:`eBayes` or :func:`classifyTestsF`, can then be applied to
    fitted model object.

    The :code:`coefficients` argument provides a simpler way to specify the
    :code:`contrasts` matrix when the desired contrasts are just a subset of
    the original coefficients.

    Note
    ----
    For efficiency reasons, this function does not re-factorize the design
    matrix for each probe. A consequence is that, if the design matrix is
    non-orthogonal and original fit included precision weights or missing
    values, then the unscaled standard deviations produced by this function are
    approximate rather than exact. The approximation is usually acceptable. If
    not, then the issue can be avoided by redefining the design matrix to fit
    the contrasts directly.

    Even with precision weights, the results from :func:`contrasts_fit` are
    always exact if the coefficients being compared are statistically
    independent. This will be true, for example, if the original fit was a
    oneway model without blocking and the group-means (no-intercept)
    parameterization was used for the design matrix.

    Arguments
    ---------
    fit : MArrayLM
        object produced by :func:`lm_series` or equivalent. Must contain
        components :code:`coefficients` and :code:`stdev_unscaled`.
    contrasts : array_like
        numeric matrix with rows corresponding to coefficients in :code:`fit`
        and columns containing contrasts. May be a vector if there is only one
        contrast. :code:`NA` are not allowed.
    coefficients : array_like
        array indicating which coefficients are to be kept in the revised fit
        object. An alternative way to specify the :code:`contrasts`.

    Returns
    -------
    MArrayLM
        an object of the same class as :code:`fit`, with components:

        - :code:`coefficients`: matrix containing the estimated coefficients
          for each contrast for each probe
        - :code:`stdev_unscaled`: matrix conformal with :code:`coefficients`
          containing the unscaled standard deviations for the coefficient
          estimators
        - :code:`cov_coefficients`: matrix giving the unscaled covariance
          matrix of the estimable coefficients

        Most other attributes of :code:`fit` are pass through unchanged, but
        :code:`t`, :code:`p_value`, :code:`lods`, :code:`F` and
        :code:`F_p_value` will all be removed.
    """
    fit = copy.deepcopy(fit)
    if (contrasts is None) == (coefficients is None):
        raise ValueError("must specify exactly one of contrasts or coefficients")

    if coefficients is not None:
        return fit[:, coefficients]

    if fit.coefficients is None:
        raise ValueError("fit must contain coefficients component")
    if fit.stdev_unscaled is None:
        raise ValueError("fit must contain stdev_unscaled component")

    # Remove test statistics in case eBayes() has previously been run on the fit object
    try:
        del fit.t
    except AttributeError:
        pass
    try:
        del fit.p_value
    except AttributeError:
        pass
    try:
        del fit.lods
    except AttributeError:
        pass
    try:
        del fit.F
    except AttributeError:
        pass
    try:
        del fit.F_p_value
    except AttributeError:
        pass

    # Number of coefficients in fit
    ncoef = fit.coefficients.shape[1]

    # Check contrasts
    if np.any(np.isnan(contrasts)):
        raise ValueError("contrasts must be a numeric matrix")
    if contrasts.ndim == 1:
        contrasts = contrasts[:, None]
    if contrasts.shape[0] != ncoef:
        raise ValueError(
            "Number of rows of contrast matrix must match number of coefficients in fit"
        )

    fit.contrasts = contrasts

    # Special case of contrast matrix with 0 columns
    if contrasts.shape[1] == 0:
        return fit[:, 0]

    # Correlation matrix of estimable coefficients
    # Test whether design was orthogonal
    if fit.cov_coefficients is None:
        logging.warnings.warn(
            "cov_coefficients not found in fit -- assuming coefficients are orthogonal"
        )
        var_coef = np.mean(fit.stdev_unscaled**2, axis=0)
        fit.cov_coefficients = np.diag(var_coef)
        cormatrix = np.eye(ncoef)
        orthog = True
    else:
        cormatrix = cov2cor(fit.cov_coefficients)
        if len(cormatrix) < 2:
            orthog = True
        else:
            orthog = np.sum(np.abs(np.tril(cormatrix, -1))) < 1e-12

    # if design matrix was singular, reduce to estimable coefficients
    r = cormatrix.shape[0]
    if r < ncoef:
        raise NotImplementedError()

    # Remove coefficients that do not appear in any contrast
    # (Not necessary but can make function faster)
    # TODO

    # Replace NA coefficients with large but finite standard deviations
    # to allow zero contrast entries to clobber NA coefficients
    i = np.isnan(fit.coefficients)
    NACoef = np.any(i)
    if NACoef:
        fit.coefficients[i] = 0
        fit.stdev_unscaled[i] = 1e30

    # New coefficients
    fit.coefficients = fit.coefficients @ contrasts

    # Test whether design was orthogonal
    if len(cormatrix) < 2:
        orthog = True
    else:
        orthog = np.all(np.abs(np.tril(cormatrix, -1)) < 1e-14)

    # New correlation matrix
    R = scipy.linalg.cholesky(fit.cov_coefficients)
    tmp = R @ contrasts
    fit.cov_coefficients = tmp.T @ tmp

    # New standard deviations
    if orthog:
        fit.stdev_unscaled = np.sqrt(fit.stdev_unscaled**2 @ contrasts**2)
    else:
        R = scipy.linalg.cholesky(cormatrix)
        if isinstance(cormatrix, pd.DataFrame):
            R = pd.DataFrame(R, index=cormatrix.index, columns=cormatrix.columns)
        ngenes = fit.stdev_unscaled.shape[0]
        ncont = contrasts.shape[1]
        U = np.ones((ngenes, ncont))
        U = pd.DataFrame(U, index=fit.stdev_unscaled.index, columns=contrasts.columns)
        o = np.ones((1, ncoef))
        for i in range(ngenes):
            RUC = R @ (contrasts.T * fit.stdev_unscaled.iloc[i, :]).T
            U.iloc[i, :] = np.sqrt(o @ RUC**2)
        fit.stdev_unscaled = U

    # replace NAs if necessary
    if NACoef:
        i = fit.stdev_unscaled > 1e20
        fit.coefficients[i] = np.nan
        fit.stdev_unscaled[i] = np.nan

    return fit



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/exactTest.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np
import pandas as pd

from ..utils import asfactor
from .DGEExact import DGEExact
from .exactTestByDeviance import exactTestByDeviance
from .exactTestBySmallP import exactTestBySmallP
from .exactTestDoubleTail import exactTestDoubleTail
from .mglmOneGroup import mglmOneGroup
from .q2qnbinom import q2qnbinom


def exactTest(
    self,
    pair=(0, 1),
    dispersion="auto",
    rejection_region="doubletail",
    big_count=900,
    prior_count=0.125,
):
    """
    Compute genewise exact tests for differences in the means between two groups of negative-binomially distributed counts.

    This function tests for differential expression between two groups of count
    libraries. It implements the exact test proposed by [Robinson2008]_ for a
    difference in mean between two groups of negative binomial random
    variables.  The functions accept two groups of count libraries, and a test
    is performed for each row of data. For each row, the test is conditional on
    the sum of counts for that row. The test can be viewed as a generalization
    of the well-known exact binomial test (implemented in :func:`binomTest`) but
    generalized to overdispersed counts.

    This function is the main user-level function, and produces an object
    containing all the necessary components for downstream analysis. It calls
    one of the low-level functions :func:`exactTestDoubleTail`,
    :func:`exactTestBetaApprox`, :func:`exactTestBySmallP` or
    :func:`exactTestByDeviance` to do the *p*-value computation. The low-level
    functions all assume that the libraries have been normalized to have the
    same size, *i.e.* to have the same expected column sum under the null
    hypothesis. :func:`exactTest` equalizes the library sizes using
    :code:`equalizeLibSizes` before calling the low-level functions.

    The functions :func:`exactTestDoubleTail`, :func:`exactTestBySmallP` and
    :func:`exactTestByDeviance` correspond to different ways to define the
    two-sided rejection region when the two groups have different numbers of
    samples.  :func:`exactTestBySmallP` implements the method of small
    probabilities as proposed by [Robinson2008]_. This method corresponds
    exactly to :func:`binomTest` as the dispersion approaches zero, but gives
    poor results when the dispersion is very large.  :func:`exactTestDoubleTail`
    computed two-sided *p*-values by doubling the smaller tail probability.
    :func:`exactTestByDeviance` uses the deviance goodness of fit statistics to
    define the rejection region, and is therefore equivalent to a conditional
    likelihood ratio test.

    Note that :code:`rejection_region="smallp"` is no longer recommended. It is
    preserved as an option only for backward compatibility with earlier
    versions of edgeR. :code:`rejection_region="deviance"` has good theoretical
    statistical properties but is relatively slow to compute.
    :code:`rejection_region="doubletail"` is just slightly more conservative
    than :code:`rejection_region="deviance"`, but is recommended because of its
    much greater speed. For general remarks on different types of rejection
    regions for exact tests, see [Gibbons1975]_.

    :func:`exactTestBetaApprox` implements an asymptotic beta distribution
    approximation to the conditional count distribution. It is called by the
    other functions for rows with both group counts greater than
    :code:`big_count`.

    Arguments
    ---------
    pair : pair of ints or of strings
        the pair of groups to be compared. If strings, then should be the names
        of two groups (*e.g.* two levels of :code:`self.samples["group"]`). If
        integers, then groups to be compared are chosen by finding the levels
        :code:`self.samples["group"]` corresponding to those indices and using
        those levels as the groups to be compared. If :code:`None`, then first
        two levels of :code:`self.samples["group"]` (a factor) are used. Note that
        the first group listed in the pair is the baseline for the comparison,
        so if the pair is :code:`("A","B")` then the comparison is :code:`B -
        A`, so genes with positive log-fold changes are up-regulated in group B
        compared with group A (and vice versa for genes with negative log-fold
        change)
    dispersion : array_like of floats, or {"auto", "common", "trended", "tagwise"}
        an array of dispersions or a string indicating that dispersions should
        be taken from the data object. If floats, then can be either of length
        one or of length equal to the number of genes. Defaults to
        :code:`"auto"` to use the most complex dispersions found in data
        object.
    rejection_region : {"doubletail", "smallp", "deviance"}
        type of rejection region for two-sided exact test.
    big_count : int
        count size above which asymptotic beta approximation will be used.
    prior_count : float
        average prior count used to shrink log-fold-changes. Larger values
        produce more shrinkage.

    Returns
    -------
    DGEExact
        dataframe with two additional components:

        - :code:`comparison`, string giving the names of the two groups being compared.
        - :code:`genes`, dataframe containing annotation for each gene; taken
          from :code:`self`

        The dataframe columns has the same rows as :code:`self` and contains
        the following columns:

        - :code:`"log2FoldChange"`, log2-fold-change of expression between
          conditions being tested.
        - :code:`"lfcSE"`, standard error of log2-fold-change.
        - :code:`"logCPM"`, average log2-counts per million.
        - :code:`"pvalue"`, the two-sided *p*-values.
    """
    # Check input
    if len(pair) != 2:
        raise ValueError("Pair must be a pair!")
    if rejection_region not in ["doubletail", "deviance", "smallp"]:
        raise ValueError(f"Invalid value {rejection_region} for 'rejection_region'")

    # Get group names
    group = asfactor(self.samples["group"])
    levs_group = group.categories
    if isinstance(pair[0], int) and isinstance(pair[1], int):
        pair = levs_group[list(pair)]

    if not all(np.isin(pair, levs_group)):
        raise ValueError(
            f"At least one element of given pair is not a group.\nGroups are {' '.join(levs_group)}"
        )

    # Get dispersion vector
    if dispersion is None:
        dispersion = "auto"
    if isinstance(dispersion, str):
        if dispersion == "common":
            dispersion = self.common_dispersion
        elif dispersion == "trended":
            dispersion = self.trended_dispersion
        elif dispersion == "tagwise":
            dispersion = self.tagwise_dispersion
        elif dispersion == "auto":
            dispersion = self.getDispersion()
        else:
            raise ValueError(f"Invalid value {dispersion} for 'dispersion'")

        if dispersion is None:
            raise ValueError("specified dispersion not found in object")
        if np.isnan(dispersion).any():
            raise ValueError("dispersion is NA")
    ntags = self.counts.shape[0]
    try:
        dispersion = np.broadcast_to(dispersion, (ntags,))
    except ValueError:
        raise ValueError(
            "Dispersion provided by user must have length either 1 or the number of tags in the DGEList object"
        )

    # Reduce to two groups
    j = np.isin(group, pair)
    y = self.counts.loc[:, j]
    lib_size = self.samples["lib_size"][j]
    norm_factors = self.samples["norm_factors"][j]
    group = group[j]
    # TODO row names

    # Normalized library sizes
    lib_size = lib_size * norm_factors
    offset = np.log(lib_size)
    lib_size_average = np.exp(offset.mean())

    # logFC
    prior_count = prior_count * lib_size / lib_size.mean()
    offset_aug = np.log(lib_size + 2 * prior_count)
    j1 = group == pair[0]
    n1 = sum(j1)
    if n1 == 0:
        raise ValueError(f"No libraries for {pair[0]}")
    y1 = y.loc[:, j1]
    abundance1 = mglmOneGroup(
        y1.to_numpy() + prior_count[j1].to_numpy(),
        offset=offset_aug[j1],
        dispersion=dispersion,
    )
    j2 = group == pair[1]
    n2 = sum(j2)
    if n2 == 0:
        raise ValueError(f"No libraries for {pair[1]}")
    y2 = y.loc[:, j2]
    abundance2 = mglmOneGroup(
        y2.to_numpy() + prior_count[j2].to_numpy(),
        offset=offset_aug[j2],
        dispersion=dispersion,
    )
    logFC = (np.asarray(abundance2) - np.asarray(abundance1)) / np.log(2)

    # following parts is inspired from DESeq2 function fitBeta
    # x is the design matrix -- here, 2 groups, so a matrix of shape (n1+n2,2)
    # fitted1 and fitted2 are the fitted values in groups 1 and 2, respectively
    # contrast is the contrast matrix between groups 1 and 2
    x = np.zeros((n1 + n2, 2))
    x[:n1, 0] = 1
    x[n1:, 1] = 1
    contrast = np.array([[-1], [1]])

    fitted1 = np.exp(abundance1[:, None] + np.asarray(offset_aug[j1]))
    wvec1 = fitted1 / (1.0 + dispersion[:, None] * fitted1)
    fitted2 = np.exp(abundance2[:, None] + np.asarray(offset_aug[j2]))
    wvec2 = fitted2 / (1.0 + dispersion[:, None] * fitted2)
    w_vec = np.hstack([wvec1, wvec2])

    ridge = np.diag(np.repeat(1e-6 / (np.log(2) ** 2), x.shape[1]))
    xtwxr_inv = np.linalg.inv(x.T @ (x * w_vec[:, :, None]) + ridge)
    assert xtwxr_inv.shape == (y.shape[0], 2, 2)
    # sigma is the covariance matrix for the logFC
    sigma = xtwxr_inv @ x.T @ (x * w_vec[:, :, None]) @ xtwxr_inv
    assert sigma.shape == (y.shape[0], 2, 2)
    lfcSE = np.sqrt(contrast.T @ sigma @ contrast) / np.log(2)
    lfcSE = lfcSE.squeeze()
    assert lfcSE.shape == logFC.shape, f"{lfcSE.shape} vs {logFC.shape}"

    # Equalize library sizes
    abundance = mglmOneGroup(y.to_numpy(), dispersion=dispersion, offset=offset)
    e = np.exp(abundance)
    input_mean = np.broadcast_to(e, (n1, ntags))
    output_mean = input_mean * lib_size_average
    input_mean = (input_mean.T * np.broadcast_to(lib_size[j1], input_mean.T.shape)).T
    y1 = q2qnbinom(
        y1.T, input_mean=input_mean, output_mean=output_mean, dispersion=dispersion
    ).T
    input_mean = np.broadcast_to(e, (n2, ntags))
    output_mean = input_mean * lib_size_average
    input_mean = (input_mean.T * np.broadcast_to(lib_size[j2], input_mean.T.shape)).T
    y2 = q2qnbinom(
        y2.T, input_mean=input_mean, output_mean=output_mean, dispersion=dispersion
    ).T

    if rejection_region == "doubletail":
        exact_pvals = exactTestDoubleTail(
            y1, y2, dispersion=dispersion, big_count=big_count
        )
    elif rejection_region == "deviance":
        exact_pvals = exactTestByDeviance(y1, y2, dispersion=dispersion)
    elif rejection_region == "smallp":
        exact_pvals = exactTestBySmallP(y1, y2, dispersion=dispersion)

    AveLogCPM = self.AveLogCPM
    if AveLogCPM is None:
        AveLogCPM = self.aveLogCPM()
    de_out = pd.DataFrame(
        {
            "log2FoldChange": logFC,
            "lfcSE": lfcSE,
            "logCPM": AveLogCPM,
            "pvalue": exact_pvals,
        },
        index=self.counts.index,
    )
    return DGEExact(table=de_out, comparison=pair, genes=self.genes)



# -----------------------------------------------------------------------------
# Copyright (C) 2024 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import numpy as np
import pandas as pd
from scipy.stats import combine_pvalues, false_discovery_control
from statsmodels.stats.meta_analysis import combine_effects

from .DEResults import DEResults


def meta_de(de_results, alpha=0.05, min_common_genes=None):
    """
    Combine logFC and *p*-values of differential expression analyses

    log-fold-changes are combined using a random-effect model, where the random
    effects variance is iteratively estimated with the Paule-Mandel method.
    Confidence intervals for the combined log-fold-change values are computed
    assuming a normal distribution and without scaling.

    *p*-values are combined using Fisher's combined probability test, then
    adjusted for multiple testing with Benjamini-Hochberg procedure.

    Arguments
    ---------
    de_results : list of DEResults
        the list of differential expression results to combine. Depending on
        the use-case, it can be results obtained with different tools on the
        same dataset, results obtained with the same tool on different
        datasets, or any combination thereof
    alpha : float between 0 and 1
        significance level for the confidence intervals. Defaults to 0.05.
    min_common_genes : int or None
        minimal number of genes all the elements of :code:`de_results` need to
        have in common. Below this threshold, an error will be raised. If
        :code:`None`, then all elements of :code:`de_results` must have the
        same set of genes.

    Returns
    -------
    pd.DataFrame
        a dataframe indexed by genes with the following columns:

        - :code:`"combined logFC"`: the combined log-fold-change
        - :code:`"combined logFC (CI_L)"`: the lower bound of the confidence
          interval for the combined log-fold-change
        - :code:`"combined logFC (CI_R)"`: the lower bound of the confidence
          interval for the combined log-fold-change
        - :code:`"adjusted combined pval"`: the combined *p*-value, adjusted
          for multiple testing
    """
    if len(de_results) < 2:
        raise ValueError("metaanalysis requires at least two analyses to aggregate")
    if not all(isinstance(d, DEResults) for d in de_results):
        raise ValueError(
            "diffexp results fed to metanalysis must inherit from DEResults"
        )

    idx = list(set.intersection(*[set(d.index) for d in de_results]))
    if min_common_genes is None:
        max_n = np.max([d.shape[0] for d in de_results])
        if len(idx) < max_n:
            raise ValueError(
                "diffexp results fed to metaanalysis must have the same set of genes"
            )
        idx = de_results[0].index
    elif len(idx) < min_common_genes:
        raise ValueError(
            f"diffexp results fed to metaanalysis must have at least {min_common_genes} in common"
        )

    de_results = [d.loc[idx, :] for d in de_results]

    # combining logFC
    logFC = {g: np.array([d.loc[g, "log2FoldChange"] for d in de_results]) for g in idx}
    lfcSE = {g: np.array([d.loc[g, "lfcSE"] for d in de_results]) for g in idx}

    ce = {g: combine_effects(logFC[g], lfcSE[g] ** 2) for g in idx}
    # NB: using random effect model with no scaling
    res = pd.DataFrame(
        {g: c.conf_int(alpha=alpha)[1] for g, c in ce.items()},
        index=["combined logFC (CI_L)", "combined logFC (CI_R)"],
    ).T
    res["combined logFC"] = [ce[g].mean_effect_re for g in res.index]

    # combining p-values
    pvals = {g: [d.loc[g, "pvalue"] for d in de_results] for g in idx}
    meta_pvals = np.array(
        [combine_pvalues(pvals[g], nan_policy="omit").pvalue for g in idx]
    )

    # multiple testing
    nan_meta_pvals = np.isnan(meta_pvals)
    meta_adj_pvals = np.zeros(meta_pvals.shape)
    meta_adj_pvals[nan_meta_pvals] = np.nan
    meta_adj_pvals[~nan_meta_pvals] = false_discovery_control(
        meta_pvals[~nan_meta_pvals]
    )
    res["adjusted combined pval"] = meta_adj_pvals

    return res



import unittest
from unittest import mock

import matplotlib.pyplot as plt
import numpy as np
import numpy.testing as npt
import pandas as pd
from sklearn.decomposition import PCA

from inmoose.cohort_qc.cohort_metric import CohortMetric


class TestCohortMetric(unittest.TestCase):
    def setUp(self):
        """Set up mock data for testing."""
        self.clinical_df = pd.DataFrame(
            {
                "SampleID": ["Sample1", "Sample2", "Sample3", "Sample4"],
                "Covariate1": [1, 1, 2, 2],
                "Covariate2": ["X", "Y", "X", "Y"],
                "Covariate3": ["D", "D", "D", "D"],
                "batch": [
                    "Dataset1",
                    "Dataset1",
                    "Dataset2",
                    "Dataset2",
                ],
            }
        ).set_index("SampleID")

        self.data_expression_df = pd.DataFrame(
            {
                "Sample1": [1.1, 1.2, 1.3],
                "Sample2": [2.1, 2.2, 2.3],
                "Sample3": [3.1, 3.2, 3.3],
                "Sample4": [4.1, 4.2, 4.3],
            },
            index=["Gene1", "Gene2", "Gene3"],
        )

        self.data_expression_df_before = self.data_expression_df.copy()

        self.covariates = ["Covariate1", "Covariate2"]

        self.qc = CohortMetric(
            clinical_df=self.clinical_df,
            batch_column="batch",
            data_expression_df=self.data_expression_df,
            data_expression_df_before=self.data_expression_df_before,
            covariates=self.covariates,
            n_components=2,
            n_neighbors=2,
        )

    def test_cohort_qc_only_one_batch(self):
        self.clinical_df["batch"] = "unique_batch"
        with self.assertRaises(ValueError):
            CohortMetric(
                clinical_df=self.clinical_df,
                batch_column="batch",
                data_expression_df=self.data_expression_df,
                data_expression_df_before=self.data_expression_df_before,
                covariates=self.covariates,
                n_components=2,
                n_neighbors=2,
            )

    def test_covariate_corner_cases(self):
        """
        make sure the code runs without error when passing zero or one covariate
        """
        CohortMetric(
            clinical_df=self.clinical_df,
            batch_column="batch",
            data_expression_df=self.data_expression_df,
            data_expression_df_before=self.data_expression_df_before,
            n_components=2,
            n_neighbors=2,
        )
        CohortMetric(
            clinical_df=self.clinical_df,
            batch_column="batch",
            data_expression_df=self.data_expression_df,
            data_expression_df_before=self.data_expression_df_before,
            covariates=[],
            n_components=2,
            n_neighbors=2,
        )
        CohortMetric(
            clinical_df=self.clinical_df,
            batch_column="batch",
            data_expression_df=self.data_expression_df,
            data_expression_df_before=self.data_expression_df_before,
            covariates=["Covariate1"],
            n_components=2,
            n_neighbors=2,
        )

    def test_CohortMetric_missing_covariates(self):
        missing_covariates = ["covariate1", "nonexistent_covariate"]

        with self.assertRaises(ValueError) as context:
            CohortMetric(
                clinical_df=self.clinical_df,
                batch_column="batch",
                data_expression_df=self.data_expression_df,
                covariates=missing_covariates,
                n_components=2,
                n_neighbors=2,
            )
        self.assertEqual(
            str(context.exception),
            "Covariates covariate1, nonexistent_covariate are not present in the clinical dataframe.",
        )

    def test_CohortMetric_without_before_data(self):
        """Test CohortMetric class."""
        qc_without_before_data = CohortMetric(
            clinical_df=self.clinical_df,
            batch_column="batch",
            data_expression_df=self.data_expression_df,
            covariates=["Covariate1", "Covariate2", "Covariate3"],
            n_components=2,
            n_neighbors=2,
        )
        assert qc_without_before_data.data_expression_df_before is None
        assert qc_without_before_data.covariates == ["Covariate1", "Covariate2"]

    def test_identify_mixed_datasets(self):
        """Test identify_mixed_datasets method."""
        mixed_datasets = self.qc.identify_mixed_datasets()
        self.assertEqual(mixed_datasets, ["Dataset1", "Dataset2"])

    @mock.patch("seaborn.scatterplot")
    def test_pca_plot(self, mock_scatterplot):
        # Simulate PCA output (2D points)
        pcs = np.random.rand(4, 2)
        labels = self.clinical_df["batch"]

        # Call the private method _pca_plot
        result_fig = self.qc.pca_plot(pcs=pcs, labels=labels, title="Test PCA")

        # Check that the method returns a matplotlib figure
        self.assertIsInstance(result_fig, plt.Figure)

        # Check that seaborn scatterplot was called
        self.assertTrue(mock_scatterplot.called)

        # Extract the actual arguments passed to the scatterplot
        _, kwargs = mock_scatterplot.call_args

        # Verify the numpy arrays pcs[:, 0] and pcs[:, 1] match
        npt.assert_array_equal(kwargs["x"], pcs[:, 0])
        npt.assert_array_equal(kwargs["y"], pcs[:, 1])

    def test_plot_pca_variance(self):
        # Example explained variance data for 5 principal components
        explained_variance = np.array([0.4, 0.3, 0.2, 0.05, 0.05])

        # Call the private method _plot_pca_variance
        result_fig = self.qc.plot_pca_variance(
            explained_variance=explained_variance, ylim=0.5
        )

        # Check that the method returns a matplotlib figure
        self.assertIsInstance(result_fig, plt.Figure)

    def test_create_correlation_matrix_with_pc(self):
        clinical_df_with_pc = self.clinical_df.copy()
        clinical_df_with_pc["Data Element With Spaces"] = [
            "Some Value",
            "other-value",
            "otherValue",
            "other-value",
        ]
        clinical_df_with_pc["PC1"] = [1, 2, 3, 1]
        clinical_df_with_pc["PC2"] = [2, 2, 3, 5]
        result_matrix = self.qc.create_correlation_matrix_with_pc(clinical_df_with_pc)

        # Pearson correlation should be done with numeric_var
        self.assertEqual(
            result_matrix.at["Covariate1", "PC1"],
            ("3.02e-01", "6.98e-01", "Pearson correlation", 4),
        )
        self.assertEqual(
            result_matrix.at["Covariate1", "PC2"],
            ("8.16e-01", "1.84e-01", "Pearson correlation", 4),
        )

        # T-test should be done for binary categorical_var
        self.assertEqual(
            result_matrix.at["Covariate2", "PC1"], ("4.47e-01", "6.98e-01", "T-test", 4)
        )
        self.assertEqual(
            result_matrix.at["Covariate2", "PC2"],
            ("-6.32e-01", "5.92e-01", "T-test", 4),
        )

    def test_pca_analysis(self):
        """Test pca_analysis method."""
        result = self.qc.pca_analysis()
        # Ensure the result is a tuple of length 6
        self.assertEqual(len(result), 6)
        self.assertIsInstance(result[0], pd.DataFrame)
        self.assertIsInstance(result[1], pd.DataFrame)
        self.assertIsInstance(result[2], PCA)
        self.assertIsInstance(result[3], PCA)
        self.assertIsInstance(result[4], np.ndarray)
        self.assertIsInstance(result[5], np.ndarray)

    def test_cohort_summary(self):
        """Test cohort_summary method."""
        summary = self.qc.cohort_summary()
        self.assertEqual(summary["Number of Samples"], 4)
        self.assertEqual(summary["Number of Datasets"], 2)
        self.assertEqual(summary["Number of Genes"], 3)

    def test_quantify_correction_effect(self):
        """Test quantify_correction_effect method."""
        mad_before, mad_after, effect_metric = self.qc.quantify_correction_effect()
        self.assertIsInstance(mad_before, float)
        self.assertIsInstance(mad_after, float)
        self.assertIsInstance(effect_metric, float)

    @mock.patch("inmoose.cohort_qc.cohort_metric.silhouette_score")
    def test_silhouette_score(self, mock_silhouette_score):
        """Test silhouette_score method."""
        mock_silhouette_score.side_effect = [0.5, 0.3]  # Mock silhouette scores
        score_before, score_after = self.qc.silhouette_score()
        self.assertEqual(score_before, 0.3)
        self.assertEqual(score_after, 0.5)

    @mock.patch("sklearn.neighbors.NearestNeighbors.kneighbors")
    def test_compute_entropy(self, mock_kneighbors):
        """Test compute_entropy method."""
        mock_kneighbors.return_value = (None, np.array([[0, 1], [1, 0]]))
        entropy = self.qc.compute_entropy(self.qc.data_expression_df)
        self.assertIsInstance(entropy, float)

    def test_entropy_batch_mixing(self):
        """Test entropy_batch_mixing method."""
        entropy_before, entropy_after = self.qc.entropy_batch_mixing()
        self.assertIsInstance(entropy_before, float)
        self.assertIsInstance(entropy_after, float)

    @mock.patch("inmoose.cohort_qc.cohort_metric.sns.violinplot")
    def test_compare_sample_distribution_by_covariates(self, mock_violinplot):
        fig = self.qc.compare_sample_distribution_by_covariates(("1", "X"))
        self.assertIsNotNone(fig)

    @mock.patch("inmoose.cohort_qc.cohort_metric.sns.violinplot")
    def test_compare_sample_distribution_total(self, mock_violinplot):
        # Test compare_sample_distribution_total
        fig = self.qc.compare_sample_distribution_total()
        self.assertIsNotNone(fig)

    def test_process(self):
        """Test process method."""
        # Mock the return values of each method that is called within `process`
        with (
            mock.patch.object(self.qc, "pca_analysis") as mock_pca_analysis,
            mock.patch.object(
                self.qc, "quantify_correction_effect"
            ) as mock_quantify_correction_effect,
            mock.patch.object(self.qc, "silhouette_score") as mock_silhouette_score,
            mock.patch.object(
                self.qc, "entropy_batch_mixing"
            ) as mock_entropy_batch_mixing,
            mock.patch.object(
                self.qc, "summarize_and_compare_mixed_datasets"
            ) as mock_summarize_and_compare_mixed_datasets,
        ):
            # Define mock return values for each method
            mock_pca_analysis.return_value = (
                pd.DataFrame(),  # association_matrix_before
                pd.DataFrame(),  # association_matrix
                PCA(),  # pca_before
                PCA(),  # pca_after
                np.array([]),  # pcs_before
                np.array([]),  # pcs_after
            )
            mock_quantify_correction_effect.return_value = (0.6, 0.4, 0.2)
            mock_silhouette_score.return_value = (0.5, 0.3)
            mock_entropy_batch_mixing.return_value = (0.7, 0.6)
            mock_summarize_and_compare_mixed_datasets.return_value = {"some": "summary"}

            # Call the process method
            self.qc.process()

            # Assert that each method was called once
            mock_pca_analysis.assert_called_once()
            mock_quantify_correction_effect.assert_called_once()
            mock_silhouette_score.assert_called_once()
            mock_entropy_batch_mixing.assert_called_once()
            mock_summarize_and_compare_mixed_datasets.assert_called_once()



import sys

import numpy
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


class build_ext_cxx17(build_ext):
    def build_extensions(self):
        std_flag = (
            "-std:c++17" if self.compiler.compiler_type == "msvc" else "-std=c++17"
        )
        for e in self.extensions:
            e.extra_compile_args.append(std_flag)
        super().build_extensions()


macros = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
profiling = False
linetrace = False
if "--profile" in sys.argv:
    profiling = True
    linetrace = True
    macros += [("CYTHON_TRACE_NOGIL", "1")]
    sys.argv.remove("--profile")

stats_cpp = Extension(
    "inmoose.utils.stats_cpp",
    [
        "inmoose/utils/_stats.pyx",
    ],
    include_dirs=[numpy.get_include()],
    define_macros=macros,
)

edgepy_cpp = Extension(
    "inmoose.edgepy.edgepy_cpp",
    [
        "inmoose/edgepy/edgepy_cpp/edgepy_cpp.pyx",
    ],
    include_dirs=[numpy.get_include()],
    define_macros=macros,
)


setup(
    cmdclass={"build_ext": build_ext_cxx17},
    packages=[
        "inmoose",
        "inmoose/consensus_clustering",
        "inmoose/data",
        "inmoose/data/airway",
        "inmoose/data/pasilla",
        "inmoose/pycombat",
        "inmoose/limma",
        "inmoose/edgepy",
        "inmoose/edgepy/edgepy_cpp",
        "inmoose/sim",
        "inmoose/utils",
        "inmoose/deseq2",
        "inmoose/diffexp",
        "inmoose/cohort_qc",
    ],
    package_data={
        "inmoose/edgepy/edgepy_cpp": [
            "edgepy_cpp.pyx",
            "__init__.pxd",
            "edgepy_cpp.h",
        ],
        "inmoose/data/airway": [
            "airway.h5ad",
        ],
        "inmoose/data/pasilla": [
            "Dmel.BDGP5.25.62.DEXSeq.chr.gff",
            "geneIDsinsubset.txt",
            "pasilla_gene_counts.tsv",
            "pasilla_sample_annotation.csv",
            "treated1fb.txt",
            "treated2fb.txt",
            "treated3fb.txt",
            "untreated1fb.txt",
            "untreated2fb.txt",
            "untreated3fb.txt",
            "untreated4fb.txt",
        ],
        "inmoose/cohort_qc": ["qc_report.html"],
    },
    ext_modules=[edgepy_cpp, stats_cpp],
)



import unittest

from inmoose.utils import round_scientific_notation, truncate_name


class TestFormat(unittest.TestCase):
    def test_round_scientific_notation(self):
        """Test rounding of small positive numbers."""
        result = round_scientific_notation(0.0000123456)
        self.assertEqual(result, "1.23e-05")

    def test_truncate_name(self):
        """Test case where the name is longer than max_length."""
        result = truncate_name("ThisNameIsTooLong", 11)
        self.assertEqual(result, "ThisName...")

        result = truncate_name("ThisNameIsShort", 15)
        self.assertEqual(result, "ThisNameIsShort")



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023-2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/plot.R' of the Bioconductor DESeq2 package
# (version 3.16).


import matplotlib.pyplot as plt
import numpy as np


def plotDispEsts_dds(
    self,
    ymin=None,
    CV=False,
    genecol="black",
    fitcol="red",
    finalcol="dodgerblue",
    legend=True,
    xlab=None,
    ylab=None,
    log="xy",
    cex=0.7,
):
    """
    A simple helper function that plots the per-gene dispersion estimates
    together with the fitted mean-dispersion relationship.

    Arguments
    ---------
    ymin
        the lower bound for points on the plot, points beyond this are drawn as
        triangles at :code:`ymin`
    CV : bool
        whether to plot the asymptotic or biological coefficient of variation
        (the square root of dispersion) on the y-axis. As the mean grows to
        infinity, the square root of dispersion gives the coefficient of
        variation for the counts. Default is :code:`False`, plotting
        dispersion.
    genecol : str
        the color for gene-wise dispersion estimates
    fitcol : str
        the color of the fitted estimates
    finalcol : str
        the color of the final estimates used for testing
    legend : bool
        whether to draw a legend
    xlab : str
    ylab : str
    log : {"", "x", "y", "xy"}
        the axis (if any) to log scale
    cex : float
        the default size of the points to plot
    """

    if xlab is None:
        xlab = "mean of normalized counts"
    if ylab is None:
        if CV:
            ylab = "coefficient of variation"
        else:
            ylab = "dispersion"

    px = self.var["baseMean"]
    sel = px > 0
    px = px[sel]

    # transformation of dispersion into CV or not
    if CV:
        f = np.sqrt
    else:

        def f(x):
            return x

    py = f(self.var["dispGeneEst"][sel])
    if ymin is None:
        ymin = 10 ** np.floor(np.log10(np.nanmin(py[py > 0])) - 0.1)

    plt.scatter(px, np.maximum(py, ymin), s=cex, c=genecol, label="gene-est")
    if "x" in log:
        plt.xscale("log")
    if "y" in log:
        plt.yscale("log")

    outliers = (self.var["dispOutlier"] == 1.0)[sel]

    if self.dispersions is not None:
        plt.scatter(
            px[~outliers],
            f(self.dispersions[sel][~outliers]),
            c=finalcol,
            s=cex,
            label="final",
        )
        # use a circle over outliers
        plt.scatter(
            px[outliers],
            f(self.dispersions[sel][outliers]),
            s=8 * cex,
            marker="o",
            facecolors="none",
            edgecolors=finalcol,
        )

    if self.var["dispFit"] is not None:
        plt.scatter(px, f(self.var["dispFit"][sel]), c=fitcol, s=cex, label="fitted")

    if legend:
        plt.legend(loc="lower right")

    plt.xlabel(xlab)
    plt.ylabel(ylab)
    plt.show()


def plotMA_dds(
    self,
    alpha=0.1,
    main="",
    xlab="mean of normalized counts",
    ylim=None,
    MLE=False,
    *args,
    **kwargs,
):
    """
    Helper function to MA-plot the results of :code:`self`
    """
    res = self.results(*args, **kwargs)
    return res.plotMA(alpha=alpha, main=main, xlab=xlab, ylim=ylim, MLE=MLE)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is a Python port of the original C++ code from the files
# 'src/R_fit_levenberg.cpp' and 'src/glm_levenberg.cpp' of the Bioconductor
# edgeR package (version 3.38.4).

import numpy as np
from scipy.linalg import LinAlgError, cho_factor, cho_solve

from .nbdev import nb_deviance

low_value = 1e-10
supremely_low_value = 1e-13
ridiculously_low_value = 1e-100


def fit_levenberg(y, offset, disp, weights, design, beta, tol, maxit):
    """
    Arguments
    ---------
    y : array_like
        matrix of counts
    offset : array_like
        offsets, same shape as :code:`y`
    disp : array_like
        dispersions, same shape as :code:`y`
    weights : array_like
        weights, same shape as :code:`y`
    design : array_like
        design matrix, as many rows as columns in :code:`y`
    beta : array_like
        initial values of the beta, as many rows as in :code:`y`, as many
        columns as in :code:`design`
    tol : float
        tolerance for convergence
    maxit : int
        maximal number of iterations

    Returns
    -------
    ndarray
        fitted beta, same shape as :code:`design`
    ndarray
        fitted mu, same shape as :code:`y`
    ndarray
        genewise deviance
    ndarray
        genewise number of iterations
    ndarray
        whether convergence was reached, genewise
    """

    assert offset.shape == y.shape
    assert disp.shape == y.shape
    assert weights.shape == y.shape
    assert design.shape[0] == y.shape[1]
    assert beta.shape == (y.shape[0], design.shape[1])

    out_beta = beta.copy()
    mu = np.zeros(y.shape)
    dev = np.zeros(y.shape[0])
    iter_ = np.zeros(y.shape[0])
    conv = np.repeat(False, y.shape[0])

    # We compute mu based on beta. Returning if there are no coeffs
    mu = np.exp(out_beta @ design.T + offset)
    dev = nb_deviance(y, mu, weights, disp)

    # for all-zero libraries, there is really no point continuing
    ymax = np.max(y, axis=1)
    out_beta[ymax < low_value, :] = np.nan
    mu[ymax < low_value, :] = 0
    conv[ymax < low_value] = False
    dev[ymax < low_value] = 0
    iter_[ymax < low_value] = 0

    remaining = ~(ymax < low_value)

    # iterating using reweighted least squares
    for i in range(maxit):
        if not remaining.any():
            break

        iter_[remaining] += 1

        # Here we set up the matrix XtWX i.e. the Fisher information matrix.
        # X is the design matrix and W is a diagonal matrix with the working
        # weights for each observation (i.e. library). The working weights are
        # part of the first derivative of the log-likelihood for a given
        # coefficient, multiplied by any user-specified weights. When multiplied
        # by two covariates in the design matrix, you get the Fisher information
        # (i.e. variance of the log-likelihood) for that pair. This takes the
        # role of the second derivative of the log-likelihood. The working
        # weights are formed by taking the reciprocal of the product of the
        # variance (in terms of the mean) and the square of the derivative of
        # the link function.

        # We also set up the actual derivative of the log likelihoods in 'dl'.
        # This is done by multiplying each covariate by the difference between
        # the mu and observation and dividing by the variance and derivative of
        # the link function. This is then summed across all observations for
        # each coefficient. The aim is to solve (XtWX)(dbeta)=dl for 'dbeta'.
        # As XtWX is the second derivative, and dl is the first, you can see
        # that we are effectively performing a multivariate Newton-Raphson
        # procedure with 'dbeta' as the step.

        denom = 1 + mu * disp
        deriv = (y - mu) / denom * weights

        working_weights = mu / denom * weights
        xtwx = (design.T * np.expand_dims(working_weights, axis=-2)) @ design

        dl = deriv @ design
        assert dl.shape == (y.shape[0], design.shape[1])
        dbeta = np.zeros(dl.shape)
        maxinfo = np.max(np.diagonal(xtwx, axis1=-1, axis2=-2), axis=1)
        if i == 0:
            lambda_ = np.maximum(maxinfo * 1e-6, supremely_low_value)

        # Levenberg/Marquardt damping reduces step size until the deviance
        # increases or no step can be found that increases the deviance. In
        # short, increases in the deviance are enforced to avoid problems with
        # convergence.
        lev = np.repeat(0, y.shape[0])
        low_dev = np.repeat(False, y.shape[0])

        lev_rem = remaining.copy()
        while lev_rem.any():
            lev[lev_rem] += 1

            for tag in range(y.shape[0]):
                if not lev_rem[tag]:
                    continue

                while True:
                    # We add lambda_ to the diagonal. This reduces the step size
                    # as the second derivative is increased.
                    xtwx_copy = xtwx[tag] + lambda_[tag] * np.eye(xtwx.shape[1])

                    # Cholesky decomposition, and then use the decomposition to
                    # solve for dbeta in (XtWX)dbeta = dl.
                    try:
                        xtwx_cho = cho_factor(xtwx_copy)
                        break
                    except LinAlgError:
                        # If it fails, it MUST mean that the matrix is singular
                        # due to numerical imprecision as all the diagonal
                        # entries of the XtWX matrix must be positive. This
                        # occurs because of fitted values being exactly zero;
                        # thus, the coefficients attempt to converge to
                        # negative infinity. This generally forces the step
                        # size to be larger (i.e. lambda_ lower) in order to
                        # get to infinity faster (which is impossible). Low
                        # lambda_ leads to numerical instability and effective
                        # singularity. To solve this, we actually increase
                        # lambda_; this avoids code breakage to give the other
                        # coefficients a chance to converge. Failure of
                        # convergence for the zero fitted values is not a
                        # problem as the change in deviance from small ->
                        # smaller coefficients is not that great when the true
                        # value is negative infinity.
                        lambda_[tag] *= 10
                        if lambda_[tag] <= 0:
                            # just to make sure it actually increases
                            lambda_[tag] = ridiculously_low_value

                dbeta[tag, :] = cho_solve(xtwx_cho, dl[tag, :])

            # Updating beta and the means. 'dbeta' stores 'Y' from the
            # solution of (X*VX)Y=dl, corresponding to a NR step.
            beta_new = np.zeros(beta.shape)
            mu_new = np.zeros(mu.shape)
            beta_new[lev_rem, :] = out_beta[lev_rem, :] + dbeta[lev_rem, :]
            mu_new[lev_rem, :] = np.exp(
                beta_new[lev_rem, :] @ design.T + offset[lev_rem, :]
            )

            # Checking if the deviance has decreased or if it is too small to
            # care about. Either case is good and means that we will be using
            # the updated fitted values and coefficients. Otherwise, if we have
            # to repeat the inner loop, then we want to do so from the original
            # values (as we will be scaling lambda up so we want to retake the
            # step from where we were before). This is why we do not modify the
            # values in-place until we are sure we want to take the step.
            dev_new = nb_deviance(y, mu_new, weights, disp)
            with np.errstate(invalid="ignore"):
                low_dev[(dev_new / ymax) < supremely_low_value] = True
            low_tags = lev_rem & ((dev_new <= dev) | low_dev)
            out_beta[low_tags, :] = beta_new[low_tags, :]
            mu[low_tags, :] = mu_new[low_tags, :]
            dev[low_tags] = dev_new[low_tags]
            lev_rem &= ~low_tags

            # Increasing lambda_, to increase damping. Again, we have to
            # make sure it is not zero.
            lambda_[lev_rem] *= 2
            lambda_[lev_rem & (lambda_ <= 0)] = ridiculously_low_value

            # Excessive damping; steps get so small that it is pointless to
            # continue.
            with np.errstate(divide="ignore"):
                conv[lev_rem & ((lambda_ / maxinfo) > (1 / supremely_low_value))] = True
            lev_rem &= ~conv

        # Terminating if we failed, if divergence from the exact solution is
        # acceptably low (cross-product of dbeta with the log-likelihood
        # derivative) or if the actual deviance of the fit is acceptably low.
        remaining &= ~conv
        remaining &= ~low_dev
        # NB: np.diag(A @ B.T) is equivalent to np.sum(A * B, axis=1)
        divergence = np.sum(dl * dbeta, axis=1)
        remaining &= ~(divergence < tol)

        # If we quit the inner Levenberg loop immediately and survived all the
        # break conditions above, that means that deviance is decreasing
        # substantially. Thus, we need larger steps to get there faster. To do
        # so, we decrease the damping factor. Note that this only applies if we
        # did not decrease the damping factor in the inner Levenberg loop, as
        # that would indicate that we need to slow down.
        lambda_[remaining & (lev == 1)] /= 10

    return (out_beta, mu, dev, iter_, conv)


# for reference, but it is likely dead code
def autofill(design, beta, offset):
    return np.exp(design @ beta + offset)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is a Python port of the original C++ code from the files
# 'src/nbdev.cpp' and 'src/R_compute_nbdev.cpp' of the Bioconductor edgeR
# package (version 3.38.4).

import numpy as np

from .edgepy_cpp import compute_unit_nb_deviance


def nb_deviance(y, mu, w, phi):
    """
    Calculate the genewise deviance of a negative binomial fit

    See also
    --------
    compute_unit_nb_deviance

    Arguments
    ---------
    y : array_like
        counts matrix
    mu : array_like
        expected means matrix (broadcastable to the shape of :code:`y`)
    w : array_like
        observation weights matrix (broadcastable to the shape of :code:`y`)
    phi : array_like
        dispersion matrix (broadcastable to the shape of :code:`y`)

    Returns
    -------
    ndarray
        matrix of deviances, with as many elements as rows in :code:`y`
    """
    return np.sum(w * compute_unit_nb_deviance(y, mu, phi), axis=1)



# -----------------------------------------------------------------------------
# Copyright (C) 2024 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import numpy as np


def cov2cor(covariance):
    """
    Scale a covariance matrix into the corresponding correlation matrix

    Arguments
    ---------
    covariance : ndarray
        the covariance matrix to scale

    Returns
    -------
    ndarray
        the corresponding correlation matrix
    """
    v = np.sqrt(np.diag(covariance))
    outer_v = np.outer(v, v)
    correlation = covariance / outer_v
    correlation[covariance == 0] = 0
    return correlation
